/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : cancel.c                                                 */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Contains functions for cancelling active timers          */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Jun 2001                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <timer.h>      /* Specific header for this dll                       */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <BipCni.h>		/* Function prototypes for Plug-in Node functions     */

/* local function prototypes */
/******************************************************************************/
/* evaluateTimerCancel                                                        */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Locate matching timer based on match specification and cancel it.          */
/*                                                                            */
/******************************************************************************/

typedef struct fullKey {
	char msgId[MQ_MSG_ID_LENGTH],
		 correlId[MQ_CORREL_ID_LENGTH];
} fullKey_t;

int evaluateTimerCancel(IN context_t* contextP,
 			   		    IN CciMessage *destinationList,
		                IN CciMessage *exceptionList,
		                IN CciMessage *message)
{
	deployedTimerCancel_t *deployedTimerCancelP;
	void *correlIdElementP,
		 *messageIdElementP;
    struct CciByteArray byteArrayDetails;
    CciSize numBytes;
	fullKey_t matchKey;
	int rc,
		retCode,
		numMatches,
		nameStringLength,
		messageNumber;
	bool_t nullCorrelId,
		   nullMsgId;
	char nullId[MQ_MSG_ID_LENGTH],
		 *nameString,
		 mbString[10 * MAX_STRING_SIZE + 1],
		 msgBuffer[1000];
    static char *functionName="evaluateTimerCancel";

    TRACE_ENTRY("evaluateTimerCancel", ("contextP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 contextP, destinationList,
							 exceptionList, message));

	/* Using the match conditions specified on the TimerCancel node, look for any
	 * matching timers and cancel them.  There may be multiple matches.
	 */

	deployedTimerCancelP = (deployedTimerCancel_t *)(contextP->nodeDataP);

	/* extract the message id and correl id values */
	messageIdElementP = getMessageElement(message, MESSAGE_ID);

	if (!messageIdElementP)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Message Id not found in message"));
		TRACE_RETURN(RSN_014, TT_MESSAGE_ID_NOT_FOUND);
	}

	/* copy message id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)matchKey.msgId;
	byteArrayDetails.size = sizeof(matchKey.msgId);
	numBytes = cniElementByteArrayValue(&rc,
										messageIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Message Id from message"));
		TRACE_RETURN(RSN_015, TT_MESSAGE_ID_NOT_FOUND);
	}

	TRACE_MID(TRACE_INFO, ("Incoming Message Id value is..."));
	TRACE_MIDHEX(TRACE_INFO, matchKey.msgId, MQ_MSG_ID_LENGTH);

	/* extract the correlId */
	correlIdElementP = getMessageElement(message, CORREL_ID);

	if (!correlIdElementP)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Correlation Id not found in message"));
		TRACE_RETURN(RSN_014, TT_CORRELID_NOT_FOUND);
	}

	/* copy correl id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)matchKey.correlId;
	byteArrayDetails.size = sizeof(matchKey.correlId);
	numBytes = cniElementByteArrayValue(&rc,
										correlIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Correl Id from message"));
		TRACE_RETURN(RSN_015, TT_CORRELID_NOT_FOUND);
	}

	TRACE_MID(TRACE_INFO, ("Incoming Correl Id value is..."));
	TRACE_MIDHEX(TRACE_INFO, matchKey.correlId, MQ_CORREL_ID_LENGTH);

	/* Check for null keys, i.e. either MsgId or CorrelId set to binary zeroes */
	nullCorrelId = FALSE;
	nullMsgId = FALSE;
	memset(nullId, 0, sizeof(nullId));
	if (!memcmp(matchKey.correlId, nullId, sizeof(nullId)))
	{
		nullCorrelId = TRUE;
        TRACE_MID(TRACE_INFO, ("CorrelId is null"));
	}
	if (!memcmp(matchKey.msgId, nullId, sizeof(nullId)))
	{
		nullMsgId = TRUE;
        TRACE_MID(TRACE_INFO, ("MsgId is null"));
	}

	TRACE_MID(TRACE_INFO, ("About to lock timerChainMutex"));
	threadMutexLock(timerChainMutex);
	TRACE_MID(TRACE_INFO, ("Locked timerChainMutex"));

	TRACE_MID(TRACE_INFO, ("deployedTimerCancelP->timerName is '%S'",
		                    deployedTimerCancelP->timerName));
	TRACE_MID(TRACE_INFO, ("deployedTimerCancelP->retrieveTimerNameFromTree is '%s'",
		deployedTimerCancelP->retrieveTimerNameFromTree ? "TRUE":"FALSE"));

	/* search required index for matching timers and cancel them */ 
	numMatches = 0;
	nameString = NULL;
	nameStringLength = 0;
	if (deployedTimerCancelP->timerName && UClen(deployedTimerCancelP->timerName))
	{
		if (deployedTimerCancelP->retrieveTimerNameFromTree)
		{
			/* locate specified element and extract its value */
			TRACE_MID(TRACE_INFO, ("About to retrieve timer name from message"));
			
			retCode = getValueFromSyntaxTree(message,
										     deployedTimerCancelP->timerName,
											 TRUE, // allow child VALUE type node to be used
											 &nameString,
											 &nameStringLength);

			if (retCode != TT_SUCCESS)
			{
				/* deal with common errors */
				if (retCode == TT_ELEMENT_NOT_FOUND)
				{
					UCsprintf(msgBuffer, "Failed to locate syntax element specified for the '%S' "
									   "attribute in message.  The element path is '%S'",
									   ATTR_TIMER_NAME_MB,
									   deployedTimerCancelP->timerName);
					messageNumber = MJB_FAILED_TO_LOCATE_SYNTAX_ELEMENT;
				}
				else
				if (retCode == TT_ELEMENT_HAS_NO_VALUE)
				{
					UCsprintf(msgBuffer, "Failed to locate any data in syntax element specified "
									   "for the '%S' attribute in message.  The element path is '%S'",
									   ATTR_TIMER_NAME_MB,
									   deployedTimerCancelP->timerName);
					messageNumber = MJB_SYNTAX_ELEMENT_CONTAINS_NO_DATA;
				}
				else
				if (retCode == TT_ELEMENT_DATA_CONTAINS_NULL_DATA_CHARACTER)
				{
					UCsprintf(msgBuffer, "Failed to extract any data from syntax element specified "
									   "for the '%S' attribute in message because the data "
									   "contains one or more null characters.  The element path "
									   "is '%S'",
									   ATTR_TIMER_NAME_MB,
									   deployedTimerCancelP->timerName);
					messageNumber = MJB_SYNTAX_ELEMENT_DATA_CONTAINS_NULLS;
				}
				else
				{
					/* some other error has occurred */
					UCsprintf(msgBuffer, "Error occurred when trying to extract data from syntax "
									   "element specified for the '%S' attribute in message.  "
									   "The element path is '%S'",
									   ATTR_TIMER_NAME_MB,
									   deployedTimerCancelP->timerName);
					messageNumber = MJB_SYNTAX_ELEMENT_DATA_RETRIEVAL_FAILURE;
				}

				/* issue the appropriate log message */
				cciLog(&rc, 
					   CCI_LOG_ERROR, 
					   __FILE__, 
					   __LINE__, 
					   functionName, 
					   C.PLUGIN_MESSAGE_CATALOGUE, 
					   messageNumber, 
					   msgBuffer,
#if !defined(WIN32)
					   brokerName,
#endif
					   ATTR_TIMER_NAME_MB,
 					   /* convert MessageData to multi byte string */
					   UCToStr(deployedTimerCancelP->timerName,
							   mbString,
							   sizeof(mbString),
							   NULL),
					   (char *)0);
									 
				TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'failure' "
									   "terminal", message));
				/* 
				 * Propagate message to the 'failure' terminal if element doesn't exist or 
				 * has no data.
				 */
				retCode = propagateMessage(contextP,
										   C.TERM_FAILURE,
										   message,
										   destinationList,
										   exceptionList,
										   FALSE); /* this is not a new message */

				if (retCode != TT_SUCCESS)
				{
					TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'failure' terminal"));
					RETHROW_IF_EXCEPTION(retCode);
					TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
				}

				TRACE_RETURN(RSN_013, TT_FAILED);
			}
		}
		else
		{
			TRACE_MID(TRACE_INFO, ("Extracting literal timer name value"));
			/* use literal value from deployed cancel node */
			nameStringLength = UClen(deployedTimerCancelP->timerName)+1;
			nameString = (char *)malloc(nameStringLength);
		    UCToStr(deployedTimerCancelP->timerName,
				    nameString,
				    nameStringLength,
				    NULL);
		}
		TRACE_MID(TRACE_INFO, ("Using %sname '%s' to search name index", 
			deployedTimerCancelP->retrieveTimerNameFromTree ? "derived " : "literal ",
			                    nameString));
		retCode = searchIndexAndCancel(nameString,
									   INDEX_NAME, 
							           deployedTimerCancelP->scope,
							           &numMatches);
	}
	else
	{
 		switch (deployedTimerCancelP->matchValue)
		{
			case MATCH_TYPE_CORRELID_MSGID:
				if (nullCorrelId)
				{
					TRACE_MID(TRACE_INFO, ("No match on null CorrelId"));
					break;
				}
				TRACE_MID(TRACE_INFO, ("Using CorrelId value to search MsgId index"));
				retCode = searchIndexAndCancel(matchKey.correlId, 
					                           INDEX_MSGID, 
										       deployedTimerCancelP->scope,
										       &numMatches);
				break;

			case MATCH_TYPE_MSGID_MSGID:
				if (nullMsgId)
				{
					TRACE_MID(TRACE_INFO, ("No match on null MsgId"));
					break;
				}
				TRACE_MID(TRACE_INFO, ("Using MsgId value to search MsgId index"));
				retCode = searchIndexAndCancel(matchKey.msgId, 
					                           INDEX_MSGID, 
										       deployedTimerCancelP->scope,
										       &numMatches);
				break;

			case MATCH_TYPE_CORRELID_CORRELID:
				if (nullCorrelId)
				{
					TRACE_MID(TRACE_INFO, ("No match on null CorrelId"));
					break;
				}
				TRACE_MID(TRACE_INFO, ("Using CorrelId value to search CorrelId index"));
				retCode = searchIndexAndCancel(matchKey.correlId, 
					                           INDEX_CORRELID, 
										       deployedTimerCancelP->scope,
										       &numMatches);
				break;

			case MATCH_TYPE_MSGID_CORRELID:
				if (nullMsgId)
				{
					TRACE_MID(TRACE_INFO, ("No match on null MsgId"));
					break;
				}
				TRACE_MID(TRACE_INFO, ("Using MsgId value to search CorrelId index"));
				retCode = searchIndexAndCancel(matchKey.msgId, 
					                           INDEX_CORRELID, 
										       deployedTimerCancelP->scope,
										       &numMatches);
				break;

			case MATCH_TYPE_BOTH:
				if (nullMsgId && nullCorrelId)
				{
					TRACE_MID(TRACE_INFO, ("No match on null MsgId AND null CorrelId"));
					break;
				}
				TRACE_MID(TRACE_INFO, ("Using MsgId+CorrelId value to search Both index"));
				retCode = searchIndexAndCancel((char *)&matchKey, 
					                           INDEX_BOTH, 
										       deployedTimerCancelP->scope,
										       &numMatches);
				break;

			default:
				// should never get here...
				TRACE_MID(TRACE_ERROR, ("Deployed timer match value is invalid (%d)",
					                     deployedTimerCancelP->matchValue));
				break;
		}
	}

	if (nameString)
	{
		free(nameString);
	}

	if (!numMatches)
	{
		TRACE_MID(TRACE_INFO, ("No matching timer found"));

		TRACE_MID(TRACE_INFO, ("About to unlock timerChainMutex"));
		threadMutexUnlock(timerChainMutex);

		/* propagate the message to the noMatch terminal */
		TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'noMatch' "
							   "terminal", message));
		retCode = propagateMessage(contextP,
								   C.TERM_NO_MATCH,
								   message,
								   destinationList,
								   exceptionList,
								   FALSE); /* this is not a new message */

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'noMatch' terminal"));
			RETHROW_IF_EXCEPTION(retCode);
			TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_019, TT_SUCCESS);
	}

	TRACE_MID(TRACE_INFO, ("About to unlock timerChainMutex"));
	threadMutexUnlock(timerChainMutex);

	/* propagate message */
	TRACE_MID(TRACE_INFO, ("Propagating message to 'out' terminal"));
	retCode = propagateMessage(contextP,
							   C.TERM_OUT,
							   message,
							   destinationList,
							   exceptionList,
						       TRUE); /* this is a new message */

	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'out' terminal"));
		RETHROW_IF_EXCEPTION(retCode);
		TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
	}

	TRACE_RETURN(RSN_019, TT_SUCCESS);
}

/******************************************************************************/
/* searchIndexAndCancel                                                       */
/******************************************************************************/
int searchIndexAndCancel(IN char *key,
				         IN indexType_t indexType,
				         IN scope_t scope,
				         OUT int *numMatchesP)
{
	int numMatches = 0,
	    keyOffset,
		keyLen,
		numSlots = 0,
		nextSlot = 0,
		i;
	bool_t firstTimeThrough,
		   removeFromGlobalChains,
		   nextInstanceSet,
		   matchingKey;
	timerInstance_t *currentInstanceP,
					*nextInstanceP = NULL;
	globalTimerInstance_t *currentGlobalInstanceP,
						  *nextGlobalInstanceP,
						  *deadAnchorP;
	scope_t currentScope;

    TRACE_ENTRY("searchIndexAndCancel", ("key=%.8x, indexType=%s, "
		                                 "scope=%s", 
							 key, 
							 indexType==INDEX_MSGID ? "INDEX_MSGID" :
							 indexType==INDEX_CORRELID ? "INDEX_CORRELID" :
							 indexType==INDEX_BOTH ? "INDEX_BOTH" : "INDEX_NAME",
							 scope == SCOPE_EXECUTION_GROUP ? "SCOPE_EXECUTION_GROUP" :
							 scope == SCOPE_BROKER ? "SCOPE_BROKER" : "SCOPE_PLATFORM"));

	/* on entry, assume caller has locked the timer chain mutex */

	/* we always search the local index first for matches */

    if (indexType == INDEX_MSGID)
	{
		currentInstanceP = msgIdIndexAnchorP->nextMsgIdP;
		keyOffset = 0;
		keyLen = MQ_MSG_ID_LENGTH;
	}
	else
    if (indexType == INDEX_CORRELID)
	{
		currentInstanceP = correlIdIndexAnchorP->nextCorrelIdP;
		keyOffset = MQ_MSG_ID_LENGTH;
		keyLen = MQ_CORREL_ID_LENGTH;
	}
	else
    if (indexType == INDEX_BOTH)
	{
		currentInstanceP = bothIndexAnchorP->nextBothP;
		keyOffset = 0;
		keyLen = MQ_MSG_ID_LENGTH + MQ_CORREL_ID_LENGTH;
	}
	else // INDEX_NAME
	{
		currentInstanceP = nameIndexAnchorP->nextNameP;
		keyOffset = 0;
		keyLen = strlen(key);
	}

	/* search timer instance chain for matching key */
	firstTimeThrough = TRUE;
	nextInstanceSet = FALSE;
	while(currentInstanceP)
	{
		if (!firstTimeThrough)
		{
			if (nextInstanceSet)
			{
				currentInstanceP = nextInstanceP;
				nextInstanceSet = FALSE;
			}
			else
			/* check next in list */
			if (indexType == INDEX_MSGID)
			{
				currentInstanceP = currentInstanceP->nextMsgIdP;
			}
			else
			if (indexType == INDEX_CORRELID)
			{
				currentInstanceP = currentInstanceP->nextCorrelIdP;
			}
			else
			if (indexType == INDEX_BOTH)
			{
				currentInstanceP = currentInstanceP->nextBothP;
			}
			else  // INDEX_NAME
			{
				currentInstanceP = currentInstanceP->nextNameP;
			}
		}
		else
		{
			firstTimeThrough = FALSE;
		}

		if (!currentInstanceP)
		{
			continue;
		}

		TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentInstanceP));
		matchingKey = FALSE;
		if (indexType == INDEX_NAME)
		{
			/* check if names match */
			if (currentInstanceP->global)
			{
				currentGlobalInstanceP = (globalTimerInstance_t *)currentInstanceP;
				if (currentGlobalInstanceP->timerName != SHM_NULL_HANDLE)
				{
					matchingKey = 
						!strcmp(SHM_HANDLE2ADDRESS(currentGlobalInstanceP->timerName, char *, currentGlobalInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform),
								key);
				}
			}
			else
			if (currentInstanceP->timerName)
			{
			    matchingKey = !strcmp(currentInstanceP->timerName, key);
			}
		}
		else
		{
		    matchingKey = !memcmp(currentInstanceP->msgId+keyOffset,
			                      key,
				                  keyLen);
		}
		if (matchingKey)
		{
			TRACE_MID(TRACE_INFO, ("Found a matching %stimer at %.8x",
									currentInstanceP->global ? "global " : "",
									currentInstanceP));

			/* address next in list, before we free up the current entry */
			if (indexType == INDEX_MSGID)
			{
				nextInstanceP = currentInstanceP->nextMsgIdP;
			}
			else
			if (indexType == INDEX_CORRELID)
			{
				nextInstanceP = currentInstanceP->nextCorrelIdP;
			}
			else
			if (indexType == INDEX_BOTH)
			{
				nextInstanceP = currentInstanceP->nextBothP;
			}
			else
			{
				nextInstanceP = currentInstanceP->nextNameP; // INDEX_NAME
			}
			nextInstanceSet = TRUE;

			/* 
			 * If its a global entry, then we need to take the global lock before we can
			 * unchain it.  This is to prevent other processes from accessing it while we
			 * are removing it.
			 */
			if (currentInstanceP->global)
			{
				/* lock the shared memory */
				shmLock(currentInstanceP->scope==SCOPE_BROKER);

				/* remove from local and global chains and free timer */
				TRACE_MID(TRACE_INFO, ("Free'ing global timer at %.8x", currentInstanceP));

				// this global timer is owned by the current process, so it needs to be removed
				// from all chains, both local and global.

				freeTimer(currentInstanceP);

				/* unlock the shared memory */
				shmUnlock(currentInstanceP->scope==SCOPE_BROKER);
			}
			else
			{
				/* remove from local and global chains and free timer */
				TRACE_MID(TRACE_INFO, ("Free'ing local timer at %.8x", currentInstanceP));
				freeTimer(currentInstanceP);
			}

			numMatches++;
		}
	}

	/* If only local indexes were being searched, then return to caller */
	if (scope==SCOPE_EXECUTION_GROUP)
	{
		/* return the match list */
		*numMatchesP = numMatches;
		TRACE_RETURN(RSN_013, TT_SUCCESS);
	}

	/*******************************************************************/
	/*******************************************************************/
	/* The scope is global - we also need to search the global chain(s)*/
	/*                                                                 */
	/* Any matching timers found during this search must be owned by   */
	/* another process since 'our' timers would have been found during */
	/* the previous search.  They are removed from all global index    */
	/* chains and added to the global dead chain instead.              */
	/*                                                                 */
    /* NB: Platform Scope is a superset of Broker Scope.               */
	/*******************************************************************/
	/*******************************************************************/
	for (i=0; i<2; i++)
	{
		// first time through, search the Broker scope chain 
		if (!i)
		{
			// if this is Platform scope, then Broker shared memory may not be attached
			if (scope == SCOPE_PLATFORM && !sharedMemoryCBP_Broker) {
			    TRACE_MID(TRACE_INFO, ("Broker scope shared memory not attached"));
				continue;
			}
			currentScope = SCOPE_BROKER;
		}
		else
		{
			// check if Platform scope chain needs to be searched also
			if (scope != SCOPE_PLATFORM)
			{
				break;
			}
			currentScope = SCOPE_PLATFORM;
		}

		/* lock shared memory */
		shmLock(currentScope==SCOPE_BROKER);

		if (indexType == INDEX_MSGID)
		{
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS((currentScope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalMsgIdP,
								   globalTimerInstance_t *, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			keyOffset = 0;
			keyLen = MQ_MSG_ID_LENGTH;
		}
		else
		if (indexType == INDEX_CORRELID)
		{
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS((currentScope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalCorrelIdP,
								   globalTimerInstance_t *, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			keyOffset = MQ_MSG_ID_LENGTH;
			keyLen = MQ_CORREL_ID_LENGTH;
		}
		else
		if (indexType == INDEX_BOTH)
		{
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS((currentScope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalBothP,
								   globalTimerInstance_t *, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			keyOffset = 0;
			keyLen = MQ_MSG_ID_LENGTH + MQ_CORREL_ID_LENGTH;
		}
		else
		{
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS((currentScope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalNameP,
								   globalTimerInstance_t *, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			keyOffset = 0;
			keyLen = strlen(key);
		}

		/* search timer instance chain for matching key */
		firstTimeThrough = TRUE;
		while(currentGlobalInstanceP)
		{
			if (!firstTimeThrough)
			{
				if (nextGlobalInstanceP)
				{
					currentGlobalInstanceP = nextGlobalInstanceP;
					nextGlobalInstanceP = NULL;
				}
				else
				/* check next in list */
				if (indexType == INDEX_MSGID)
				{
					currentGlobalInstanceP = 
						SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalMsgIdP,
										   globalTimerInstance_t *, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
				}
				else
				if (indexType == INDEX_CORRELID)
				{
					currentGlobalInstanceP = 
						SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalCorrelIdP,
										   globalTimerInstance_t *, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
				}
				else
				if (indexType == INDEX_BOTH)
				{
					currentGlobalInstanceP = 
						SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalBothP,
										   globalTimerInstance_t *, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
				}
				else
				{
					currentGlobalInstanceP = 
						SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalNameP,
										   globalTimerInstance_t *, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
				}
			}
			else
			{
				firstTimeThrough = FALSE;
			}

			if (!currentGlobalInstanceP)
			{
				continue;
			}

			TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentGlobalInstanceP));

			removeFromGlobalChains = FALSE;

			matchingKey = FALSE;
			if (indexType == INDEX_NAME)
			{
				/* check if names match */
				if (currentGlobalInstanceP->timerName != SHM_NULL_HANDLE)
				{
					matchingKey = !strcmp(
						SHM_HANDLE2ADDRESS(currentGlobalInstanceP->timerName, char *, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform), key);
				}
			}
			else
			{
				matchingKey = !memcmp(currentGlobalInstanceP->msgId+keyOffset,
									  key,
									  keyLen);
			}
			if (matchingKey)
			{
				TRACE_MID(TRACE_INFO, ("Found a matching global timer at %.8x",
										currentGlobalInstanceP));

				/* address next global instance before we unchain current one */
				if (indexType == INDEX_MSGID)
				{
					nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalMsgIdP, globalTimerInstance_t *, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
				}
				else
				if (indexType == INDEX_CORRELID)
				{
					nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalCorrelIdP, globalTimerInstance_t *, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
				}
				else
				if (indexType == INDEX_BOTH)
				{
					nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalBothP, globalTimerInstance_t *, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
				}
				else
				{
					nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalNameP, globalTimerInstance_t *, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
				}
				
				unchainTimer((timerInstance_t *)currentGlobalInstanceP); 

				/* set the timer to expired */
				currentGlobalInstanceP->expired = TRUE;
				TRACE_MID(TRACE_INFO, ("Setting timer expired to TRUE"));

				/* add to deadInstanceAnchor */
				deadAnchorP = &(currentScope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalDeadInstanceAnchor;

				INSERT_AFTER_IN_GLOBAL_TIMER_CHAIN(deadAnchorP, currentGlobalInstanceP, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
				/* mark this timer as being on the dead chain */
				currentGlobalInstanceP->onDeadChain = TRUE;
				TRACE_MID(TRACE_INFO, ("Timer added to dead chain"));

				numMatches++;
			}
		}

		/* unlock shared memory */
		shmUnlock(currentScope=SCOPE_BROKER);
	}

	/* return the match list */
	*numMatchesP = numMatches;
	TRACE_RETURN(RSN_014, TT_SUCCESS);
}

/******************************************************************************/
/* freeTimer                                                                 */
/******************************************************************************/
void freeTimer(IN timerInstance_t *timerInstanceP)
{
	globalTimerInstance_t *globalTimerInstanceP;
    TRACE_ENTRY("freeTimer", ("timerInstanceP=%.8x", timerInstanceP));

	// This function may be called to deal with a number of different timer contexts...
	// 
	// Local Timer
	//		- remove from all local chains and free
	//
	// Locally owned global timer, expired by alarm thread
	//		- remove from all global and local chains and free globally
	//
	// Locally owned global timer, cancelled by another process
	//		- remove from all local chains and dead chain and free globally
	//

	if (timerInstanceP->global)	{
		globalTimerInstanceP = (globalTimerInstance_t *)timerInstanceP;

		// If this is on the dead chain, then it means another process has cancelled the
		// timer and removed it from all other global chains
		if (globalTimerInstanceP->onDeadChain) {  // is it on the dead chain?
			TRACE_MID(TRACE_INFO, ("Processing global timer cancelled by another process"));
			
			REMOVE_FROM_GLOBAL_TIMER_CHAIN(globalTimerInstanceP, globalTimerInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			TRACE_MID(TRACE_INFO, ("Global timer removed from dead chain"));

			/* remove timer from local chains */
			unchainTimer(timerInstanceP);
		}
		else {
			// This is a global timer expired by our own alarm thread or cancelled by our own
			// process
			TRACE_MID(TRACE_INFO, ("Processing global timer expired or cancelled by this process"));

			// remove timer from both local and global chains
			unchainTimer(timerInstanceP);  
		}

		shmFree(timerInstanceP, TRUE, timerInstanceP->scope==SCOPE_BROKER);
		TRACE_MID(TRACE_INFO, ("global timerInstanceP free'd"));
	}
	else {
		TRACE_MID(TRACE_INFO, ("Processing local timer"));

		// process local timer (non-global)
		unchainTimer(timerInstanceP);

		/* free up the instance control block */
		if (timerInstanceP->savedDataP)
		{
			TRACE_MID(TRACE_INFO, ("About to free savedDataP"));
			free(timerInstanceP->savedDataP);
		}
		if (timerInstanceP->timerName)
		{
			TRACE_MID(TRACE_INFO, ("About to free timerName"));
			free(timerInstanceP->timerName);
		}

		free(timerInstanceP);
		TRACE_MID(TRACE_INFO, ("timerInstanceP free'd"));
	}

    TRACE_RETURN_VOID(RSN_001);
}

/******************************************************************************/
/* unchainTimer                                                              */
/******************************************************************************/
void unchainTimer(IN timerInstance_t *timerInstanceP)
{
	bool_t timerOwnedByThisProcess = TRUE;
	globalTimerInstance_t *globalTimerInstanceP;
    TRACE_ENTRY("unchainTimer", ("timerInstanceP=%.8x", 
		                          timerInstanceP));

	/* assume caller has appropriate locks held on entry... */

	/* check if this process owns the timer */
	if (timerInstanceP->global) {
		globalTimerInstanceP = (globalTimerInstance_t *)timerInstanceP;
		if (globalTimerInstanceP->executionGroupPID != currentPID) {
			timerOwnedByThisProcess = FALSE;
			TRACE_MID(TRACE_INFO, ("Timer is owned by foreign process (%d)", 
									globalTimerInstanceP->executionGroupPID));
		}
	}

	/* remove from all local chains if we own this timer */
	if (timerOwnedByThisProcess)
	{
		/* unchain timer instance */
		REMOVE_FROM_LINKED_LIST(timerInstanceP);
		TRACE_MID(TRACE_INFO, ("Removed from timer instance chain"));

		/* only remove from alarm chain if interval value was specified */
		if (timerInstanceP->deployedTimerP->intervalSeconds != INFINITE)
		{
			REMOVE_FROM_ALARM_CHAIN(timerInstanceP);
			TRACE_MID(TRACE_INFO, ("Removed from alarm chain"));
		}

		if (!timerInstanceP->global && timerInstanceP->timerName ||
			 timerInstanceP->global && (shmemHandle_t)timerInstanceP->timerName != SHM_NULL_HANDLE)
		{
			/* we only need to remove this entry from the name chain */
			REMOVE_FROM_NAME_CHAIN(timerInstanceP);
			TRACE_MID(TRACE_INFO, ("Removed from name chain"));
		}
		else
		{
			REMOVE_FROM_MSG_ID_CHAIN(timerInstanceP);
			TRACE_MID(TRACE_INFO, ("Removed from MsgId chain"));

			REMOVE_FROM_CORREL_ID_CHAIN(timerInstanceP);
			TRACE_MID(TRACE_INFO, ("Removed from CorrelId chain"));

			REMOVE_FROM_BOTH_CHAIN(timerInstanceP);
			TRACE_MID(TRACE_INFO, ("Removed from Both chain"));
		}
	}

	/* if timer is a global instance, remove from relevant global chains */
	if (timerInstanceP->global) {

		globalTimerInstanceP = (globalTimerInstance_t *)timerInstanceP;

		/* if its on the dead chain, then do nothing */
		if (globalTimerInstanceP->onDeadChain) {
			TRACE_MID(TRACE_INFO, ("Timer on dead chain - no global chains removed"));
			TRACE_RETURN_VOID(RSN_002);
		}

		/* assume caller has got shared memory lock */
		if (globalTimerInstanceP->timerName != SHM_NULL_HANDLE)
		{
			/* we only need to remove this entry from the global name chain */
			REMOVE_FROM_GLOBAL_TIMER_NAME_CHAIN(globalTimerInstanceP, globalTimerInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			TRACE_MID(TRACE_INFO, ("Removed from global name chain"));
		}
		else
		{
			REMOVE_FROM_GLOBAL_TIMER_MSG_ID_CHAIN(globalTimerInstanceP, globalTimerInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			TRACE_MID(TRACE_INFO, ("Removed from global MsgId chain"));

			REMOVE_FROM_GLOBAL_TIMER_CORREL_ID_CHAIN(globalTimerInstanceP, globalTimerInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			TRACE_MID(TRACE_INFO, ("Removed from global CorrelId chain"));

			REMOVE_FROM_GLOBAL_TIMER_BOTH_CHAIN(globalTimerInstanceP, globalTimerInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			TRACE_MID(TRACE_INFO, ("Removed from global Both chain"));
		}
	}

    TRACE_RETURN_VOID(RSN_002);
}

/******************************************************************************/
/* applyDataToElement                                                         */
/******************************************************************************/
int applyDataToElement(IN CciElement *elementP, 
					   IN timerInstance_t *timerP)
{
	char *dataBuffer,
		 *currentPos,
		 *endPos;
	int retCode,
		dataBufferLength,
		depth = 0;
	CciElement *localElementP = elementP;
	bool_t moreDataInBuffer,
		   firstTime = TRUE;
    TRACE_ENTRY("applyDataToElement", ("elementP=%.8x, timerP=%.8x", 
		                                elementP, timerP));

	if (timerP->global)
	{
		TRACE_MID(TRACE_INFO, ("Global timer being processed"));
	}
	else
	{
		TRACE_MID(TRACE_INFO, ("Local timer being processed"));
	}

	dataBuffer = timerP->savedDataP;
	currentPos = dataBuffer;
	dataBufferLength = timerP->savedDataLength;

	if (dataBufferLength)
	{
	    moreDataInBuffer = TRUE;
		endPos = dataBuffer + dataBufferLength;
	}
	else
	{
	    moreDataInBuffer = FALSE;
	}

	while(moreDataInBuffer)
	{

		retCode = inflateSyntaxElement(&localElementP,
									   dataBuffer,
									   dataBufferLength,
									   &currentPos,
									   &depth,
									   firstTime,
									   FALSE); /* don't create element for first saved element */
		if (currentPos == endPos)
		{
			moreDataInBuffer = FALSE;
		}
		firstTime = FALSE;
	}

    TRACE_RETURN(RSN_001, TT_SUCCESS);
}


/* local functions */


