#define GLOBAL_TIMER_DATA 1
/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : timer.c                                                  */
/* Output name     : timer.dll                                                */
/*                   (NOTE: must be renamed to timer.lil and moved to         */
/*                   <install dir>\bin for use with MQSI V2.0 broker          */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\include\plugin        */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : This pair of plug-in nodes create and cancel a timer.    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Jun 2001                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include "timer.h" /* Specific header for this dll */                       
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "plugin.h"   /* common plugin node functions */
#include <BipCni.h>	  /* Function prototypes for Plug-in Node functions     */

/* local function prototypes */
int attachSharedMemoryIfRequired(IN context_t *contextP,
								 IN scope_t scope);
executionGroupEntry_t *findExecutionGroupEntry(IN bool_t scopeIsBroker);
void removeEGTimersFromGlobalStorage(IN executionGroupEntry_t *currentEGP,
									 IN bool_t scopeIsBroker);
executionGroupEntry_t *createNewExecutionGroupEntry(IN bool_t scopeIsBroker);
void removeExecutionGroupFromGlobalStorage(IN bool_t scopeIsBroker);


/******************************************************************************/
/* Node Implementation Functions                                              */
/*                                                                            */
/* These are mandatory and MUST be implemented by the programmer.             */
/*                                                                            */
/* createNodeContext                                                          */
/* deleteNodeContext                                                          */
/* evaluate                                                                   */
/* getAttribute                                                               */
/* getAttributeName                                                           */
/* setAttribute                                                               */
/* bipGetMessageflowNodeFactory                                               */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/* createNodeContext                                                          */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  factoryObject (input) - address of factory object that owns the node      */
/*                          being created.                                    */
/*  nodeName      (input) - name of node being created.                       */
/*  nodeObject    (input) - address of the node object that has just been     */
/*                          created.                                          */
/* Returns:                                                                   */
/*                                                                            */
/*  Pointer to the local context.                                             */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Creates any plug-in context for an instance of the node object. This is    */
/* invoked by the broker whenever a node instance is constructed. Nodes are   */
/* constructed whenever a messageflow containing the node is deployed by the  */
/* broker.                                                                    */
/*                                                                            */
/* This function has 4 responsibilities:                                      */
/*                                                                            */
/* 1. Optionally verify the name of the node is supported by the factory.     */
/* 2. Allocate node instance specific data areas. e.g. For attribute data or  */
/*    terminals.                                                              */
/* 3. Perform any additional resource acquisition or initialisation that is   */
/*    required for the processing of the node.                                */
/* 4. Return the address of the context to the calling function.              */
/*                                                                            */
/******************************************************************************/
CciContext *createNodeContext(CciFactory *factoryObject, 
                              CciChar *nodeName, 
                              CciNode *nodeObject)
{
    int rc;
	static char *functionName="createNodeContext()";
    context_t *contextP = NULL;
#if defined(POST_201_VERSION)
	CNI_BROKER_INFO_ST brokerInfo={CNI_BROKER_INFO_ST_DEFAULT};
#else
    CNI_BROKER_INFO_ST brokerInfo;
#endif
	CciChar messageFlowNameUnicode[MAX_STRING_SIZE + 1];
	char messageFlowNameMB[MAX_STRING_SIZE + 1];
	messageFlow_t *mfP,
	              defaultMsgFlow = {MESSAGE_FLOW_DEFAULT},
		          *messageFlowAnchorP;
	deployedTimer_t *deployedTimerP;
	deployedTimerCancel_t *deployedTimerCancelP;
	char msgBuffer[200],
		 nodeNameMB[100];

    TRACE_ENTRY("createNodeContext", ("factoryObject=%.8x, nodeName=%S, "
		                              "nodeObject=%.8x",
									  factoryObject,
									  nodeName,
									  nodeObject));

	/* initialise the broker info structure */
	ALLOCATE_STRING_STORAGE(brokerInfo.brokerName, 300);
	ALLOCATE_STRING_STORAGE(brokerInfo.executionGroupName, 300);
	ALLOCATE_STRING_STORAGE(brokerInfo.messageFlowName, 300);
	ALLOCATE_STRING_STORAGE(brokerInfo.queueManagerName, 300);

	/* get the name of our queue manager */
    TRACE_MID(TRACE_INFO, ("About to call cniGetBrokerInfo"));

	cniGetBrokerInfo(0,           /* retcode does not appear to be set */
			         nodeObject,
					 &brokerInfo);

	UCsprintf(messageFlowNameMB, "%S", brokerInfo.messageFlowName.buffer);
	UCcpy(messageFlowNameUnicode, brokerInfo.messageFlowName.buffer);
    TRACE_MID(TRACE_INFO, ("Extracted messageFlowName is '%s'",
			               messageFlowNameMB));

	/* Is this the first timer for this Message Flow */
	threadMutexLock(timerChainMutex);
	mfP = findMessageFlow(messageFlowNameUnicode, &messageFlowAnchor);

	if (!mfP)
	{
		/* create new message flow entry */
		mfP = (messageFlow_t *)malloc(sizeof(messageFlow_t));
		memcpy(mfP, &defaultMsgFlow, sizeof(messageFlow_t));
		mfP->name = UCdup(messageFlowNameUnicode);

		/* add to message flow chain */
		messageFlowAnchorP = &messageFlowAnchor;
		INSERT_AFTER_IN_LINKED_LIST(messageFlowAnchorP, mfP);
	}
	threadMutexUnlock(timerChainMutex);

    TRACE_MID(TRACE_INFO, ("Current messageFlow entry at %.8x", mfP));

	if (!brokerInfoObtained)
	{
		/* copy the process specific information, both as single and Unicode strings */
		UCsprintf(brokerName, "%S", brokerInfo.brokerName.buffer);
		UCcpy(brokerNameUC, brokerInfo.brokerName.buffer);
        TRACE_MID(TRACE_INFO, ("Extracted brokerName is '%s'",
			                   brokerName));

		UCsprintf(executionGroupName, "%S", brokerInfo.executionGroupName.buffer);
		UCcpy(executionGroupNameUC, brokerInfo.executionGroupName.buffer);
        TRACE_MID(TRACE_INFO, ("Extracted executionGroupName is '%s'",
			                   executionGroupName));

		UCsprintf(queueManagerName, "%S", brokerInfo.queueManagerName.buffer);
		UCcpy(queueManagerNameUC, brokerInfo.queueManagerName.buffer);
        TRACE_MID(TRACE_INFO, ("Extracted queueManagerName is '%s'",
			                   queueManagerName));

		brokerInfoObtained = TRUE;
	}

	FREE_STRING_STORAGE(brokerInfo.brokerName);
	FREE_STRING_STORAGE(brokerInfo.executionGroupName);
	FREE_STRING_STORAGE(brokerInfo.messageFlowName);
	FREE_STRING_STORAGE(brokerInfo.queueManagerName);

    /* Allocate a pointer to the local context */
    contextP = malloc(sizeof(context_t));

    /* If malloc was successful */
    if (!contextP)
	{
        /* Error handling if malloc failed */
        TRACE_RETURN_TYPE(RSN_022, void *, contextP);
	}

    TRACE_MID(TRACE_INFO, ("Context structure allocated at %.8x", contextP));

    /* initialise returned storage for context anchor */
    memset (contextP, 0, sizeof(context_t)); /* Clear the context area */
    contextP->nodeObject = nodeObject;
	contextP->messageFlowName = mfP->name;

    TRACE_MID(TRACE_INFO, ("<===========================>"));
    TRACE_MID(TRACE_INFO, ("%S node current", nodeName));
    TRACE_MID(TRACE_INFO, ("<===========================>"));

    /* Verify the name of the node is supported by the factory */
    if (!UCcmp(nodeName, C.TIMER_CREATE_NODE))
	{
		/* timer save node */
		contextP->nodeType = TIMER_CREATE;
        contextP->nodeName = UCdup(nodeName);

		/* Build attribute list for the node */
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_TIMER_NAME, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_MESSAGE_DATA, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_MESSAGE_DATA_TYPE, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_INTERVAL_SECONDS, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_TIMER_POP_COUNT, CCI_VALUE_TYPE_INTEGER);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_TIMER_MSG_EXPIRY, CCI_VALUE_TYPE_INTEGER);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_TIMER_SCOPE, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_TIMER_QUEUE, CCI_VALUE_TYPE_BYTEARRAY);

		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_LABEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_USER_TRACE_LEVEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_TRACE_LEVEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_USER_TRACE_FILTER, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_TRACE_FILTER, CCI_VALUE_TYPE_BYTEARRAY);

		/* this node has a single input and 2 output terminals */
		insertInputTerminalListEntry(contextP, C.TERM_IN);

		insertOutputTerminalListEntry(contextP, C.TERM_OUT);
		insertOutputTerminalListEntry(contextP, C.TERM_FAILURE);

		/* create a new deployed timer control block */
		deployedTimerP = (deployedTimer_t *)
			                         malloc(sizeof(deployedTimer_t));
		memset(deployedTimerP, 0, sizeof(deployedTimer_t));

		/* lock the timer mutex and add to global chain */
		threadMutexLock(timerChainMutex);
		INSERT_AFTER_IN_LINKED_LIST(timerAnchorP,
					                deployedTimerP);
		contextP->nodeDataP = deployedTimerP;

		/* increment the global node counter */
		numTimerCreateNodes++;
	    TRACE_MID(TRACE_INFO, ("Global count of TimerCreate nodes incremented to %d",
			                   numTimerCreateNodes));
		threadMutexUnlock(timerChainMutex);

		/* As of MQSI V2.0.2, an attribute for which the user has accepted the default, is
		 * no longer included in the deployment message.  We therefore need to initialise
		 * all attributes to their default values within the plug-in.
		 * The exception to this is a value for which no 'value=xxx' label is specified and 
		 * for which 'valueMandatory=true' has been specified.  In this case, the control 
		 * centre will force the user to enter a value into the field, and this will be 
		 * included in the deployment message.  Of course, we must still check for invalid
		 * values such as blanks.
		 */
	    TRACE_MID(TRACE_INFO, ("Setting default values"));
		setAttribute(contextP,
		             C.ATTR_TIMER_NAME,
                     C.ATTR_TIMER_NAME_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_MESSAGE_DATA,
                     C.ATTR_MESSAGE_DATA_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_MESSAGE_DATA_TYPE,
                     C.ATTR_MESSAGE_DATA_TYPE_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_INTERVAL_SECONDS,
                     C.ATTR_INTERVAL_SECONDS_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_TIMER_POP_COUNT,
                     C.ATTR_TIMER_POP_COUNT_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_TIMER_MSG_EXPIRY,
                     C.ATTR_TIMER_MSG_EXPIRY_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_TIMER_SCOPE,
                     C.ATTR_TIMER_SCOPE_DEFAULT_VALUE);
	    TRACE_MID(TRACE_INFO, ("<***********> Default values set"));
	}
	else
    if (!UCcmp(nodeName, C.TIMER_CANCEL_NODE))
	{
		/* timer cancel node */
		contextP->nodeType = TIMER_CANCEL;
        contextP->nodeName = UCdup(nodeName);

		/* Build attribute list for the node */
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_TIMER_NAME, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_MATCH_ON, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_TIMER_SCOPE, CCI_VALUE_TYPE_BYTEARRAY);

		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_LABEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_USER_TRACE_LEVEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_TRACE_LEVEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_USER_TRACE_FILTER, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_TRACE_FILTER, CCI_VALUE_TYPE_BYTEARRAY);

		/* this node has a single input terminal and 4 output terminals */
		insertInputTerminalListEntry(contextP, C.TERM_IN);

		insertOutputTerminalListEntry(contextP, C.TERM_OUT);
		insertOutputTerminalListEntry(contextP, C.TERM_FAILURE);
		insertOutputTerminalListEntry(contextP, C.TERM_NO_MATCH);

		/* create a new deployed timer cancel control block */
		deployedTimerCancelP = (deployedTimerCancel_t *)
			                         malloc(sizeof(deployedTimerCancel_t));
		memset(deployedTimerCancelP, 0, sizeof(deployedTimerCancel_t));

		/* lock the timer mutex and add to global chain */
		threadMutexLock(timerChainMutex);
		INSERT_AFTER_IN_LINKED_LIST(timerCancelAnchorP,
					                deployedTimerCancelP);

		contextP->nodeDataP = deployedTimerCancelP;

		/* increment the global node counter */
		numTimerCancelNodes++;
	    TRACE_MID(TRACE_INFO, ("Global count of TimerCancel nodes incremented to %d",
			                   numTimerCancelNodes));
		threadMutexUnlock(timerChainMutex);

		/* As of MQSI V2.0.2, an attribute for which the user has accepted the default, is
		 * no longer included in the deployment message.  We therefore need to initialise
		 * all attributes to their default values within the plug-in.
		 * The exception to this is a value for which no 'value=xxx' label is specified and 
		 * for which 'valueMandatory=true' has been specified.  In this case, the control 
		 * centre will force the user to enter a value into the field, and this will be 
		 * included in the deployment message.  Of course, we must still check for invalid
		 * values such as blanks.
		 */
	    TRACE_MID(TRACE_INFO, ("Setting default values"));
		setAttribute(contextP,
		             C.ATTR_TIMER_NAME,
                     C.ATTR_TIMER_NAME_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_MATCH_ON,
                     C.ATTR_MATCH_ON_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_TIMER_SCOPE,
                     C.ATTR_TIMER_SCOPE_DEFAULT_VALUE);
	    TRACE_MID(TRACE_INFO, ("<***********> Default values set"));
	}
	else
	{
	    /* Error handling if name of node is not supported by factory */
        TRACE_MID(TRACE_ERROR, ("Unrecognised node"));
		/* throw an exception here...*/
		UCsprintf(msgBuffer, "Unrecognised node ('%S') passed to createContext()",
			               contextP->nodeName);
		UCsprintf(nodeNameMB, "'%S'", contextP->nodeName);
		cciThrowException(&rc, 
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_UNRECOGNISED_NODE_CREATE_CONTEXT,  /* messageNumber */ 
			              msgBuffer,              /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
						  nodeNameMB,
			              (char *)0);
	    free(contextP);
	}

	/* 
	 * If this is the first timer node encountered, create the
	 * alarm thread
	 */
	if (!alarmThreadCreated)
	{
		rc = createAlarmThread(contextP);

		if (rc != TT_SUCCESS)
		{
			TRACE_RETURN_TYPE(RSN_005, void *, contextP);
		}
	}

    /* Return the address of the context to the calling function */
    TRACE_RETURN_TYPE(RSN_008, void *, contextP);
}

/******************************************************************************/
/* createAlarmThread                                                          */
/******************************************************************************/
int createAlarmThread(IN context_t *contextP)
{
	char errorMessage[1000];
	static char *functionName="createAlarmThread()";
	int rc;

    TRACE_ENTRY("createAlarmThread", ("contextP=%.8x",
		                              contextP));

    /* 
	 * During message flow initialisation, it's unlikely that this code
	 * will be run concurrently.  However, just to be sure, we'll 
	 * serialise the code through this section.
	 */
	TRACE_MID(TRACE_INFO, ("About to lock alarmThreadSyncMutex"));
	threadMutexLock(alarmThreadSyncMutex);

	/* check again in case we lost the race... */
	if (alarmThreadCreated)
	{
	 	TRACE_MID(TRACE_INFO, ("Lost the race to create alarm thread"));
		TRACE_MID(TRACE_INFO, ("About to unlock alarmThreadSyncMutex"));
		threadMutexUnlock(alarmThreadSyncMutex);
		TRACE_RETURN(RSN_005, TT_SUCCESS);
	}
		
	/* create the alarm thread */
	TRACE_MID(TRACE_INFO, ("About to create alarm thread"));
	if (threadCreateThread((threadFunction_t)processAlarms,
						  (threadParameter_t)contextP,
						   &alarmThreadHandle) != THREAD_OK)
	{
		/* need an exception here */
		TRACE_MID(TRACE_INFO, ("About to unlock alarmThreadSyncMutex"));
		threadMutexUnlock(alarmThreadSyncMutex);
		TRACE_RETURN(RSN_006, TT_FAILED);
	}
	TRACE_MID(TRACE_INFO, ("Alarm thread (%d) successfully created",
						   threadGetThreadId(alarmThreadHandle)));

	/* wait up to 10 seconds for child thread to signal us that he's up and running */
	TRACE_MID(TRACE_INFO, ("About to wait for alarm thread to initialise"));
	
	rc = threadCondTimedWait(alarmThreadSyncCondVar, 10000);

	if (rc != THREAD_OK && rc != THREAD_TIMED_OUT)
	{
		/* need an exception here */
		TRACE_MID(TRACE_INFO, ("About to unlock alarmThreadSyncMutex"));
		threadMutexUnlock(alarmThreadSyncMutex);
		TRACE_RETURN(RSN_007, TT_FAILED);
	}

	/* check response from alarm thread */
	if (!alarmThreadInitialised)
	{
		sprintf(errorMessage, "Alarm thread failed to initialise.  Reason='%s'",
							  alarmThreadErrorMessage);
		TRACE_MID(TRACE_ERROR, ("%s", errorMessage));

		/* Output a log message */
		cciLog(&rc,                       /* returnCode */
			   CCI_LOG_ERROR,       /* type */
			   __FILE__,                  /* file */
			   __LINE__,                  /* line */
			   functionName,              /* function */
			   C.PLUGIN_MESSAGE_CATALOGUE,    /* messageSource */
			   MJB_ALARM_THREAD_INIT_FAILED, /* messageNumber */ 
			   errorMessage,              /* traceText */
#if !defined(WIN32)
			   brokerName,
#endif
			   errorMessage,
			   (char *)0);
	}
	else
	{
		alarmThreadCreated = TRUE;

		TRACE_MID(TRACE_INFO, ("Alarm thread initialised ok"));
	}

	TRACE_MID(TRACE_INFO, ("About to unlock alarmThreadSyncMutex"));
	threadMutexUnlock(alarmThreadSyncMutex);
	TRACE_RETURN(RSN_008, TT_SUCCESS);
}

/******************************************************************************/
/* createSharedMemoryCleanupThread                                            */
/******************************************************************************/
int createSharedMemoryGarbageCollectionThread(IN context_t *contextP)
{
	char errorMessage[1000];
	static char *functionName="createSharedMemoryGarbageCollectionThread()";
	int rc;

    TRACE_ENTRY("createSharedMemoryGarbageCollectionThread", ("contextP=%.8x",
		                                                       contextP));

    /* 
	 * During message flow initialisation, it's unlikely that this code
	 * will be run concurrently.  However, just to be sure, we'll 
	 * serialise the code through this section.
	 */
	TRACE_MID(TRACE_INFO, ("About to lock alarmThreadSyncMutex"));
	threadMutexLock(shmemThreadSyncMutex);

	/* check again in case we lost the race... */
	if (shmemThreadCreated)
	{
	 	TRACE_MID(TRACE_INFO, ("Lost the race to create shared memory garbage collection thread"));
		TRACE_MID(TRACE_INFO, ("About to unlock shmemThreadSyncMutex"));
		threadMutexUnlock(shmemThreadSyncMutex);
		TRACE_RETURN(RSN_005, TT_SUCCESS);
	}
		
	/* create the garbage collection thread */
	TRACE_MID(TRACE_INFO, ("About to create shared memory garbage collection thread"));
	if (threadCreateThread((threadFunction_t)sharedMemoryGarbageCollection,
						  (threadParameter_t)contextP,
						   &shmemThreadHandle) != THREAD_OK)
	{
		/* need an exception here */
		TRACE_MID(TRACE_INFO, ("About to unlock shmemThreadSyncMutex"));
		threadMutexUnlock(shmemThreadSyncMutex);
		TRACE_RETURN(RSN_006, TT_FAILED);
	}
	TRACE_MID(TRACE_INFO, ("Garbage collection thread (%d) successfully created",
						   threadGetThreadId(shmemThreadHandle)));

	/* wait up to 10 seconds for child thread to signal us that he's up and running */
	TRACE_MID(TRACE_INFO, ("About to wait for garbage collection thread to initialise"));
	
	rc = threadCondTimedWait(shmemThreadSyncCondVar, 10000);

	if (rc != THREAD_OK && rc != THREAD_TIMED_OUT)
	{
		/* need an exception here */
		TRACE_MID(TRACE_INFO, ("About to unlock shmemThreadSyncMutex"));
		threadMutexUnlock(shmemThreadSyncMutex);
		TRACE_RETURN(RSN_007, TT_FAILED);
	}

	/* check response from alarm thread */
	if (!shmemThreadInitialised)
	{
		sprintf(errorMessage, "Garbage collection thread failed to initialise.  Reason='%s'",
							  shmemThreadErrorMessage);
		TRACE_MID(TRACE_ERROR, ("%s", errorMessage));

		/* Output a log message */
		cciLog(&rc,                       /* returnCode */
			   CCI_LOG_ERROR,       /* type */
			   __FILE__,                  /* file */
			   __LINE__,                  /* line */
			   functionName,              /* function */
			   C.PLUGIN_MESSAGE_CATALOGUE,    /* messageSource */
			   MJB_SHARED_MEMORY_GARBAGE_COLLECTION_THREAD_INIT_FAILED, /* messageNumber */ 
			   errorMessage,              /* traceText */
#if !defined(WIN32)
			   brokerName,
#endif
			   errorMessage,
			   (char *)0);
	}
	else
	{
		shmemThreadCreated = TRUE;

		TRACE_MID(TRACE_INFO, ("Shared memory garbage collection thread initialised ok"));
	}

	TRACE_MID(TRACE_INFO, ("About to unlock shmemThreadSyncMutex"));
	threadMutexUnlock(shmemThreadSyncMutex);
	TRACE_RETURN(RSN_008, TT_SUCCESS);
}

/******************************************************************************/
/* deleteNodeContext                                                          */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input)  - pointer to the current context                   */
/*                          created.                                          */
/* Returns:                                                                   */
/*                                                                            */
/*  Nothing.                                                                  */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Called whenever an instance of a node object is destroyed.  Any resources  */
/* aquired by the node e.g. context or other resources, should be released    */
/* at this time.                                                              */
/*                                                                            */
/******************************************************************************/
void deleteNodeContext(CciContext *dataP)
{
	context_t *contextP = (context_t *)dataP;
	int rc;
	bool_t scopeIsBroker;
    static char *functionName="deleteNodeContext()";
	deployedTimer_t *deployedTimerP;
	deployedTimerCancel_t *deployedTimerCancelP;
	timerInstance_t *timerInstanceP,
		                   *nextInstanceP;
	char infoMessage[1000],
		 *nodeNameP,
	     mbLabelString[MAX_STRING_SIZE + 1];

    TRACE_ENTRY("deleteNodeContext", ("dataP=%.8x", dataP));

	/* lock mutex just to be sure there's no race */
	TRACE_MID(TRACE_INFO, ("About to lock timerChainMutex"));
	threadMutexLock(timerChainMutex);

	/* delete deployed node structure pointed to by this context */
	if (contextP->nodeType == TIMER_CREATE)
	{
		deployedTimerP = (deployedTimer_t *)contextP->nodeDataP;
		TRACE_MID(TRACE_INFO, ("Processing deployed timer at %.8x", deployedTimerP));

		/* delete all timer instance belonging to this deployed timer */
        timerInstanceP = deployedTimerP->instanceAnchor.nextP;

		while (timerInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing timer instance at %.8x", timerInstanceP));
			nextInstanceP = timerInstanceP->nextP;

			freeTimer(timerInstanceP);

			timerInstanceP = nextInstanceP;
		}

		/* remove deployed timer from global chain */
		REMOVE_FROM_LINKED_LIST(deployedTimerP);

		/* free any linked resources */
		if (deployedTimerP->messageData)
		{
			TRACE_MID(TRACE_INFO, ("Freeing deployed timer message data"));
			free(deployedTimerP->messageData);
		}
		if (deployedTimerP->timerName)
		{
			TRACE_MID(TRACE_INFO, ("Freeing deployed timer name"));
			free(deployedTimerP->timerName);
		}

		/* finally, free the deployed node control block itself */
		free(deployedTimerP);
		TRACE_MID(TRACE_INFO, ("Freeing deployed timer at %.8x", deployedTimerP));

		/* decrement number of create nodes */
		numTimerCreateNodes--;
		TRACE_MID(TRACE_INFO, ("Number of timer create nodes decremented to %d",
			                   numTimerCreateNodes));
	}
	else
	{
		deployedTimerCancelP = (deployedTimerCancel_t *)contextP->nodeDataP;
		TRACE_MID(TRACE_INFO, ("Processing deployed timer cancel at %.8x", deployedTimerCancelP));

		/* remove deployed timer from global chain */
		REMOVE_FROM_LINKED_LIST(deployedTimerCancelP);

		/* free any linked resources */
		if (deployedTimerCancelP->timerName)
		{
			TRACE_MID(TRACE_INFO, ("Freeing deployed timer name"));
			free(deployedTimerCancelP->timerName);
		}

		/* finally, free the deployed node control block itself */
		free(deployedTimerCancelP);
		TRACE_MID(TRACE_INFO, ("Freeing deployed timer at %.8x", deployedTimerCancelP));

		numTimerCancelNodes--;
		TRACE_MID(TRACE_INFO, ("Number of timer cancel nodes decremented to %d",
			                   numTimerCancelNodes));
	}

	/* shutdown the alarm thread if there's no more active timer nodes */
	if (!numTimerCreateNodes && !numTimerCancelNodes &&
	    !alarmThreadShuttingDown && alarmThreadInitialised)
	{
		alarmThreadShuttingDown = TRUE;

		/* shutdown the alarm thread */
		shutdownAlarmRequested = TRUE;
		TRACE_MID(TRACE_INFO, ("Signalling alarm thread to shutdown"));
		threadCondSignal(alarmThreadWorkCondVar);
		TRACE_MID(TRACE_INFO, ("About to unlock timerChainMutex"));
		threadMutexUnlock(timerChainMutex);

		/* wait for alarm to terminate before proceeding */
		rc = threadJoin(alarmThreadHandle);

		TRACE_MID(TRACE_INFO, ("About to lock timerChainMutex"));
		threadMutexLock(timerChainMutex);
		if (rc == THREAD_OK)
		{
			shutdownAlarmRequested = FALSE;
			alarmThreadCreated = FALSE;
			alarmThreadInitialised = FALSE;
			alarmThreadHandle = 0;
			alarmThreadShuttingDown = FALSE;
		}
	}

	/* shutdown the garbage collector thread if there's no more active timer nodes */
	if (!numTimerCreateNodes && !numTimerCancelNodes &&
	    !shmemThreadShuttingDown && shmemThreadInitialised)
	{
		shmemThreadShuttingDown = TRUE;

		/* shutdown the garbage collector thread */
		shutdownShmemRequested = TRUE;
		TRACE_MID(TRACE_INFO, ("Signalling garbage collection thread to shutdown"));

		TRACE_MID(TRACE_INFO, ("About to lock shmemThreadWorkMutex"));
		threadMutexLock(shmemThreadWorkMutex);
		threadCondSignal(shmemThreadWorkCondVar);
		TRACE_MID(TRACE_INFO, ("About to unlock shmemThreadWorkMutex"));
		threadMutexUnlock(shmemThreadWorkMutex);

		/* wait for alarm to terminate before proceeding */
		rc = threadJoin(shmemThreadHandle);

		if (rc == THREAD_OK)
		{
			shutdownShmemRequested = FALSE;
			shmemThreadCreated = FALSE;
			shmemThreadInitialised = FALSE;
			shmemThreadHandle = 0;
			shmemThreadShuttingDown = FALSE;
		}
	}

	/* detach shared memory if attached and no more active timer nodes */
	if (!numTimerCreateNodes && !numTimerCancelNodes && 
		(sharedMemoryAttached_Broker || sharedMemoryAttached_Platform))
	{
		TRACE_MID(TRACE_INFO, ("About to lock sharedMemoryInitTermMutex"));
		threadMutexLock(shmemInitTermMutex);

		if (sharedMemoryAttached_Broker)
		{
			scopeIsBroker = TRUE;
			/* remove our execution group entry first... */
			shmLock(scopeIsBroker);
			removeExecutionGroupFromGlobalStorage(scopeIsBroker);
			shmUnlock(scopeIsBroker);
			rc = shmDetach(currentPID, scopeIsBroker);
			if (rc != TT_SUCCESS)
			{
				/* deal with errors... */
			}
			sharedMemoryAttached_Broker = FALSE;
			sharedMemoryCBP_Broker = NULL;
		}

		if (sharedMemoryAttached_Platform)
		{
			scopeIsBroker = FALSE;
			/* remove our execution group entry first... */
			shmLock(scopeIsBroker);
			removeExecutionGroupFromGlobalStorage(scopeIsBroker);
			shmUnlock(scopeIsBroker);
			rc = shmDetach(currentPID, scopeIsBroker);
			if (rc != TT_SUCCESS)
			{
				/* deal with errors... */
			}
			sharedMemoryAttached_Platform = FALSE;
			sharedMemoryCBP_Platform = NULL;
		}

		
		TRACE_MID(TRACE_INFO, ("About to unlock sharedMemoryInitTermMutex"));
		threadMutexUnlock(shmemInitTermMutex);
	}

	TRACE_MID(TRACE_INFO, ("About to unlock timerChainMutex"));
	threadMutexUnlock(timerChainMutex);

	/* clear the attributes associated with this context */
    deleteAttributeTableEntries(contextP);

	/* clear the terminals associated with this context */
    deleteAllTerminalListEntries(contextP);

	/* 
	 * Finally, free up the context structure, but first create multibyte strings for
	 * inclusion in log message.
	 */
    UCToStr(contextP->nodeLabel,
		    mbLabelString,
		    sizeof(mbLabelString),
		    NULL);

	if (contextP->nodeType == TIMER_CREATE)
	{
		nodeNameP = TIMER_CREATE_NODE_MB;
	}
	else
	{
		nodeNameP = TIMER_CANCEL_NODE_MB;
	}

	free(contextP->nodeName);
	free(contextP->nodeLabel);
    free(contextP);

	sprintf(infoMessage, "Successfully deleted node context for %s (%s)", 
		                  nodeNameP, mbLabelString);
	TRACE_MID(TRACE_INFO, ("%s", infoMessage));

    cciLog(&rc,                       /* returnCode */
	       CCI_LOG_INFORMATION,       /* type */
		   __FILE__,                  /* file */
		   __LINE__,                  /* line */
		   functionName,              /* function */
		   C.PLUGIN_MESSAGE_CATALOGUE, /* messageSource */
           MJB_NODE_CONTEXT_DELETED,  /* messageNumber */ 
		   infoMessage, /* traceText */
#if !defined(WIN32)
		   brokerName,
#endif
           nodeNameP,
		   mbLabelString,
		   (char *)0);

	TRACE_RETURN_VOID(RSN_009);
}


/******************************************************************************/
/* setAttribute                                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context        (input)  - pointer to the current context                  */
/*  attributeName  (input)  - attribute name for which value is to be set     */
/*  attributeValue (input) -  value to be set for the attribute               */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown attribute name                                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Sets the value of an attribute on a specific node instance. It is invoked  */
/* by the broker when a configuration request is received that attempts to    */
/* set the value of a node attribute. The broker will have verified that the  */
/* attribute name is valid for the node.                                      */
/*                                                                            */
/* This function has 3 responsibilities:                                      */
/*                                                                            */
/* 1. Verify value of attribute is correctly specified, otherwise a           */
/*    configuration exception should be thrown.                               */
/* 2. Store the attribute value in the context.                               */
/* 3. Throw a configuration exception if errors occur.                        */
/*                                                                            */
/******************************************************************************/
int setAttribute(CciContext *dataP,
				 CciChar *attributeName,
				 CciChar *attributeValue)
{
	int rc;
	context_t *contextP = (context_t *)dataP;

    TRACE_ENTRY("setAttribute", ("dataP=%.8x, attributeName=%S, "
		                         "attributeValue='%S'",
								 dataP, attributeName, attributeValue));

	/* call common plugin function */
	rc = commonSetAttribute(contextP,
		                    attributeName,
						    attributeValue);

	if (rc != CCI_SUCCESS)
	{
        TRACE_RETURN_TYPE(RSN_011, int, rc);
	}

	/* special processing for timer attributes */
    rc = processAttribute(contextP,
		                  attributeName,
		                  attributeValue);
	if (rc != TT_SUCCESS)
	{
	    TRACE_RETURN_TYPE(RSN_012, int, CCI_EXCEPTION_CONFIGURATION);
	}
    TRACE_RETURN_TYPE(RSN_012, int, TT_SUCCESS);
}


/******************************************************************************/
/* processAttribute                                                           */
/******************************************************************************/
int processAttribute(IN context_t *contextP,
				     IN CciChar *attributeName,
				     IN CciChar *attributeValue)
{
	deployedTimer_t *deployedTimerP;
	deployedTimerCancel_t *deployedTimerCancelP;
    CciChar copyString[MAX_STRING_SIZE + 1],
            matchString[MAX_STRING_SIZE + 1],
            scopeString[MAX_STRING_SIZE + 1],
	        queueName[MQ_Q_NAME_LENGTH + 1];
	int copyStringLength,
		matchStringLength,
		scopeStringLength,
		intervalSeconds,
		timerPopCount,
		msgExpiry,
		retCode,
		queueLength;
	static char *functionName = "processAttribute";

    TRACE_ENTRY("processAttribute", ("contextP=%.8x, attributeName=%S, "
		                             "attributeValue='%S'",
								     contextP, attributeName, attributeValue));

	/* 
	 * During message flow initialisation, it's unlikely that this code
	 * will be run concurrently.  However, just to be sure, we'll 
	 * serialise the code through this section.
	 */
	TRACE_MID(TRACE_INFO, ("About to lock timerChainMutex"));
	threadMutexLock(timerChainMutex);

	if (contextP->nodeType == TIMER_CREATE)
	{
		deployedTimerP = (deployedTimer_t *)(contextP->nodeDataP);

		/* initialise the deployed timer structure */

		/* deal with timerName */
		if (!UCcmp(attributeName, C.ATTR_TIMER_NAME))
		{
			copyStringLength = sizeof(copyString);
			if (getStringAttribute(contextP, 
					               C.ATTR_TIMER_NAME,
								   &copyStringLength,
					               copyString) == TT_SUCCESS)
			{
				/* timer name is an optional property, so check for null string */
				if (UClen(copyString))
				{
					/* free default value first */
					if (deployedTimerP->timerName)
					{
						free(deployedTimerP->timerName);
					}

					/* 
					 * The timer name can be either a literal or a reference to a syntax
					 * element.  If its the latter, we assume it begins with 'Root.'
					 */
					if (UClen(copyString) > 5 && !UCncmp(copyString, C.ROOT_DOT, UClen(C.ROOT_DOT)))
					{
						TRACE_MID(TRACE_INFO, ("Timer name value interpreted as a syntax element path"));
                        deployedTimerP->retrieveTimerNameFromTree = TRUE;
					}
					else
					{
						TRACE_MID(TRACE_INFO, ("Timer name value interpreted as a literal value"));
                        deployedTimerP->retrieveTimerNameFromTree = FALSE;
					}

					/* name was supplied so copy string into deployed timer */
					deployedTimerP->timerName = UCdup(copyString);
				}
			}
		}

		/* process the TimerQueue */
		if (!UCcmp(attributeName, C.ATTR_TIMER_QUEUE))
		{
			queueLength = sizeof(queueName);
			if (getStringAttribute(contextP, 
								   C.ATTR_TIMER_QUEUE,
								   &queueLength,
								   queueName) == TT_SUCCESS)
			{
				UCcpy(deployedTimerP->queueName, queueName);
				TRACE_MID(TRACE_INFO, ("%S value set", C.ATTR_TIMER_QUEUE));
			}
		}

		/* process the MessageData property */
		if (!UCcmp(attributeName, C.ATTR_MESSAGE_DATA))
		{
			copyStringLength = sizeof(copyString);
			if (getStringAttribute(contextP, 
					               C.ATTR_MESSAGE_DATA,
								   &copyStringLength,
					               copyString) == TT_SUCCESS)
			{
				/* 
				 * This value can be either a literal, i.e. the data itself, or it can be the 
				 * name of a syntax element from which to copy the data.
				 *
				 * If the string begins with 'Root.' then we assume the latter.
				 */

			    if (!UClen(copyString))
				{
					TRACE_MID(TRACE_INFO, ("No MessageData specified"));
				}
				else
				{
					if (UClen(copyString) == 4 && !UCcmp(copyString, C.ROOT) ||
						UClen(copyString) > 5 && !UCncmp(copyString, C.ROOT_DOT, UClen(C.ROOT_DOT)))
					{
						TRACE_MID(TRACE_INFO, ("Value interpreted as a syntax element path"));
                        deployedTimerP->retrieveMessageDataFromTree = TRUE;
					}
					else
					{
						TRACE_MID(TRACE_INFO, ("Value interpreted as a literal value"));
                        deployedTimerP->retrieveMessageDataFromTree = FALSE;
					}

					/* copy string into deployed timer, freeing any default value first */
					if (deployedTimerP->messageData)
					{
						free(deployedTimerP->messageData);
					}
					deployedTimerP->messageData = UCdup(copyString);
				}
			}
		}

		/* process the TimerScope property */
		if (!UCcmp(attributeName, C.ATTR_TIMER_SCOPE))
		{
			scopeStringLength = sizeof(scopeString);
			if (getStringAttribute(contextP, 
					               C.ATTR_TIMER_SCOPE,
								   &scopeStringLength,
					               scopeString) == TT_SUCCESS)
			{
				/* set enum value according to specified match on string */
				if (!UCcmp(scopeString, C.SCOPE_VALUE_EG))
				{
					deployedTimerP->scope = SCOPE_EXECUTION_GROUP;
			        TRACE_MID(TRACE_INFO, ("scope set to SCOPE_EXECUTION_GROUP"));
				}
				else
				{
					if (!UCcmp(scopeString, C.SCOPE_VALUE_BROKER))
					{
						deployedTimerP->scope = SCOPE_BROKER;
						TRACE_MID(TRACE_INFO, ("scope set to SCOPE_BROKER"));
					}
					else
					{
						deployedTimerP->scope = SCOPE_PLATFORM;
						TRACE_MID(TRACE_INFO, ("scope set to SCOPE_PLATFORM"));
					}

					retCode = attachSharedMemoryIfRequired(contextP,
						                                   deployedTimerP->scope);
					if (retCode != TT_SUCCESS)
					{
						/* raise an exception */
						TRACE_MID(TRACE_ERROR, ("Failed to attach shared memory segment"));
						/* generate an exception here... */
						cciThrowException(0,
										  CCI_FATAL_EXCEPTION,
										  __FILE__,
										  __LINE__,
										  functionName,
										  C.PLUGIN_MESSAGE_CATALOGUE,
										  MJB_FAILED_TO_ATTACH_SHARED_MEMORY,  /* messageNumber */ 
										  "Failed to attach shared memory segment", /* traceText */
#if !defined(WIN32)
										  brokerName,
#endif
										  (char *)0);

					}
				}
			}
		}

		/* process the IntervalSeconds property */
		if (!UCcmp(attributeName, C.ATTR_INTERVAL_SECONDS))
		{
			copyStringLength = sizeof(copyString);
			if (getStringAttribute(contextP, 
					               C.ATTR_INTERVAL_SECONDS,
								   &copyStringLength,
					               copyString) == TT_SUCCESS)
			{
				/* check if this references a syntax element */
			    if (!UClen(copyString))
				{
					TRACE_MID(TRACE_INFO, ("No %S value specified", C.ATTR_INTERVAL_SECONDS));
					// mandatory value, raise exception
				}
				else
				{
					if (UClen(copyString) == 4 && !UCcmp(copyString, C.ROOT) ||
						UClen(copyString) > 5 && !UCncmp(copyString, C.ROOT_DOT, UClen(C.ROOT_DOT)))
					{
						TRACE_MID(TRACE_INFO, ("Value interpreted as a syntax element path"));
                        deployedTimerP->retrieveIntervalSecondsFromTree = TRUE;

						/* copy string into deployed timer, freeing any default value first */
						if (deployedTimerP->intervalSecondsElement)
						{
							free(deployedTimerP->intervalSecondsElement);
						}
						deployedTimerP->intervalSecondsElement = UCdup(copyString);
							
					}
					else
					{
						TRACE_MID(TRACE_INFO, ("Value interpreted as a literal value"));

						if (!UCisNumeric(copyString))
						{
							// raise exception
						}

						/* convert to integer value */
                    	intervalSeconds = UCToInt(copyString);

						if (!intervalSeconds)
						{
							deployedTimerP->intervalSeconds = INFINITE;
							TRACE_MID(TRACE_INFO, ("deployed timer interval set to INFINTE"));
						}
						else
						{
							deployedTimerP->intervalSeconds = intervalSeconds;
							TRACE_MID(TRACE_INFO, ("deployed timer interval set to %d",
								                    deployedTimerP->intervalSeconds));
						}
					}
				}
			}
		}

		if (!UCcmp(attributeName, C.ATTR_TIMER_POP_COUNT))
		{
			if (getIntegerAttribute(contextP, 
					                C.ATTR_TIMER_POP_COUNT,
					                &timerPopCount) == TT_SUCCESS)
			/* copy maxTimerPopCount value into deployed timer - 0=infinite */
			if (!timerPopCount)
			{
				deployedTimerP->maxTimerPopCount = INFINITE;
			    TRACE_MID(TRACE_INFO, ("deployed timer pop count set to INFINTE"));
			}
			else
			{
				deployedTimerP->maxTimerPopCount = timerPopCount;
			}
		}

		if (!UCcmp(attributeName, C.ATTR_TIMER_MSG_EXPIRY))
		{
			if (getIntegerAttribute(contextP, 
					                C.ATTR_TIMER_MSG_EXPIRY,
					                &msgExpiry) == TT_SUCCESS)
			/* copy expiry time value into deployed timer - 0=infinite */
			if (!msgExpiry)
			{
				deployedTimerP->msgExpiry = INFINITE;
			    TRACE_MID(TRACE_INFO, ("deployed timer message expiry set to INFINTE"));
			}
			else
			{
				deployedTimerP->msgExpiry = msgExpiry;
			    TRACE_MID(TRACE_INFO, ("deployed timer message expiry set to %d seconds",
					                    deployedTimerP->msgExpiry));
			}
		}
	}
	else
	if (contextP->nodeType == TIMER_CANCEL)
	{
		deployedTimerCancelP = (deployedTimerCancel_t *)(contextP->nodeDataP);

		/* initialise the deployed timer cancel structure */

		/* deal with timerName */
		if (!UCcmp(attributeName, C.ATTR_TIMER_NAME))
		{
			copyStringLength = sizeof(copyString);
			if (getStringAttribute(contextP, 
					               C.ATTR_TIMER_NAME,
								   &copyStringLength,
					               copyString) == TT_SUCCESS)
			{
				/* timer name is an optional property, so check for null string */
				if (UClen(copyString))
				{
					/* free default value first */
					if (deployedTimerCancelP->timerName)
					{
						free(deployedTimerCancelP->timerName);
					}

					/* 
					 * The timer name can be either a literal or a reference to a syntax
					 * element.  If its the latter, we assume it begins with 'Root.'
					 */
					if (UClen(copyString) > 5 && !UCncmp(copyString, C.ROOT_DOT, UClen(C.ROOT_DOT)))
					{
						TRACE_MID(TRACE_INFO, ("Timer name value interpreted as a syntax element path"));
                        deployedTimerCancelP->retrieveTimerNameFromTree = TRUE;
					}
					else
					{
						TRACE_MID(TRACE_INFO, ("Timer name value interpreted as a literal value"));
                        deployedTimerCancelP->retrieveTimerNameFromTree = FALSE;
					}
					/* name was supplied so copy string into deployed timer cancel */
					deployedTimerCancelP->timerName = UCdup(copyString);
				}
			}
		}

		/* process the TimerScope property */
		if (!UCcmp(attributeName, C.ATTR_TIMER_SCOPE))
		{
			scopeStringLength = sizeof(scopeString);
			if (getStringAttribute(contextP, 
					               C.ATTR_TIMER_SCOPE,
								   &scopeStringLength,
					               scopeString) == TT_SUCCESS)
			{
				/* set enum value according to specified match on string */
				
				if (!UCcmp(scopeString, C.SCOPE_VALUE_EG))
				{
					deployedTimerCancelP->scope = SCOPE_EXECUTION_GROUP;
			        TRACE_MID(TRACE_INFO, ("scope set to SCOPE_EXECUTION_GROUP"));
				}
				else
				{
					if (!UCcmp(scopeString, C.SCOPE_VALUE_BROKER))
					{
						deployedTimerCancelP->scope = SCOPE_BROKER;
						TRACE_MID(TRACE_INFO, ("scope set to SCOPE_BROKER"));
					}
					else
					{
						deployedTimerCancelP->scope = SCOPE_PLATFORM;
						TRACE_MID(TRACE_INFO, ("scope set to SCOPE_PLATFORM"));
					}

					retCode = attachSharedMemoryIfRequired(contextP,
														   deployedTimerCancelP->scope);
					if (retCode != TT_SUCCESS)
					{
						/* raise an exception */
						TRACE_MID(TRACE_ERROR, ("Failed to attach shared memory segment"));
						/* generate an exception here... */
						cciThrowException(0,
										  CCI_FATAL_EXCEPTION,
										  __FILE__,
										  __LINE__,
										  functionName,
										  C.PLUGIN_MESSAGE_CATALOGUE,
										  MJB_FAILED_TO_ATTACH_SHARED_MEMORY,  /* messageNumber */ 
										  "Failed to attach shared memory segment", /* traceText */
#if !defined(WIN32)
										  brokerName,
#endif
										  (char *)0);

					}
				}
			}
		}

		if (!UCcmp(attributeName, C.ATTR_MATCH_ON))
		{
			matchStringLength = sizeof(matchString);
			if (getStringAttribute(contextP, 
					               C.ATTR_MATCH_ON,
								   &matchStringLength,
					               matchString) == TT_SUCCESS)
			{

				/* set enum value according to specified match on string */
				if (!UCcmp(matchString, C.MATCH_CORRELID_TO_MSGID))
				{
					deployedTimerCancelP->matchValue = MATCH_TYPE_CORRELID_MSGID;
			        TRACE_MID(TRACE_INFO, ("matchValue set to MATCH_TYPE_CORRELID_MSGID"));
				}
				else
				if (!UCcmp(matchString, C.MATCH_MSGID_TO_MSGID))
				{
					deployedTimerCancelP->matchValue = MATCH_TYPE_MSGID_MSGID;
			        TRACE_MID(TRACE_INFO, ("matchValue set to MATCH_TYPE_MSGID_MSGID"));
				}
				else
				if (!UCcmp(matchString, C.MATCH_CORRELID_TO_CORRELID))
				{
					deployedTimerCancelP->matchValue = MATCH_TYPE_CORRELID_CORRELID;
			        TRACE_MID(TRACE_INFO, ("matchValue set to MATCH_TYPE_CORRELID_CORRELID"));
				}
				else
				if (!UCcmp(matchString, C.MATCH_MSGID_TO_CORRELID))
				{
					deployedTimerCancelP->matchValue = MATCH_TYPE_MSGID_CORRELID;
			        TRACE_MID(TRACE_INFO, ("matchValue set to MATCH_TYPE_MSGID_CORRELID"));
				}
				else
				{
					deployedTimerCancelP->matchValue = MATCH_TYPE_BOTH;
			        TRACE_MID(TRACE_INFO, ("matchValue set to MATCH_TYPE_BOTH"));
				}
			}
		}
	}

	TRACE_MID(TRACE_INFO, ("About to unlock timerChainMutex"));
	threadMutexUnlock(timerChainMutex);
								  
    TRACE_RETURN(RSN_013, TT_SUCCESS);
}

/******************************************************************************/
/* evaluate                                                                   */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Performs node processing. This function contains the main processing logic */
/* of the message flow node. It is invoked by the broker when a message is    */
/* received on one of the input terminals of an instance of a node object.    */
/*                                                                            */
/* This function has 4 responsibilities:                                      */
/*                                                                            */
/* 1. Process the message in accordance with the values of any attribute on   */
/*    the node instance.                                                      */
/* 2. Process the message based on content if required.                       */
/* 3. Propagate the message to the appropriate output terminal                */
/* 4. Throw an exception if errors occur.                                     */
/*                                                                            */
/******************************************************************************/
void evaluate(CciContext *dataP,
		      CciMessage *destinationList,
		      CciMessage *exceptionList,
		      CciMessage *message)
{
	context_t *contextP;
	int rc;
	static char *functionName = "evaluate";
	char msgBuffer[200],
		 nodeName[100];

    TRACE_ENTRY("evaluate", ("dataP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 dataP, destinationList,
							 exceptionList, message));

	contextP = (context_t *)dataP;

	/* determine which function to call based on node name */
    TRACE_MID(TRACE_INFO, ("nodeName='%S'", contextP->nodeName));
	if (!UCcmp(contextP->nodeName, C.TIMER_CREATE_NODE))
	{
        TRACE_MID(TRACE_INFO, (">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"));
        TRACE_MID(TRACE_INFO, (">>>>>>>>>>> TimerCreate >>>>>>>>>>>>"));
        TRACE_MID(TRACE_INFO, (">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"));
		rc = evaluateTimerCreate(contextP,
			                      destinationList,
							      exceptionList,
							      message);
	}
	else
	if (!UCcmp(contextP->nodeName, C.TIMER_CANCEL_NODE))
	{
        TRACE_MID(TRACE_INFO, ("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<"));
        TRACE_MID(TRACE_INFO, ("<<<<<<<<<<< TimerCancel <<<<<<<<<<<<<"));
        TRACE_MID(TRACE_INFO, ("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<"));
		rc = evaluateTimerCancel(contextP,
			                     destinationList,
							     exceptionList,
							     message);
	}
	else
	{
        TRACE_MID(TRACE_ERROR, ("Unrecognised node"));
		/* throw an exception here...*/
		UCsprintf(msgBuffer, "Unrecognised node ('%S') passed to evaluate()",
			               contextP->nodeName);
		UCsprintf(nodeName, "'%S'", contextP->nodeName);
		cciThrowException(&rc, 
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_UNRECOGNISED_NODE,  /* messageNumber */ 
			              msgBuffer,              /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
						  nodeName,
			              (char *)0);
	}

	// check response from evaluate subroutines and throw exception if needed.

    TRACE_RETURN_VOID(RSN_016);
}

/******************************************************************************/
/* bipGetMessageflowNodeFactory                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  Address of factory object.                                                */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* This function is called when the message broker loads the "lil" during the */
/* initialization of the broker engine                                        */
/*                                                                            */
/* This function has 5 responsibilities:                                      */
/*                                                                            */
/* 1. create the node factory (or factories) which the "lil" supports, using  */
/*    the createNodeFactory() function, saving the returned pointer to the    */
/*    node factory for use at runtime.                                        */
/* 2. define each node supported by the (or each) factory) and pass a pointer */
/*    to a function table which contains pointers to the node implementation  */
/*    functions.                                                              */
/* 3. define the name of each attribute supported by each node, including a   */
/*    pointer to a validation function.                                       */
/* 4. define input and output terminals supported by each node.               */
/* 5. return the pointer to the factory, unless an error occurred, in which   */
/*    case zero should be returned.                                           */    
/*                                                                            */
/******************************************************************************/
#if defined(WIN32)
    #include <process.h>
#elif defined(_AIX) || defined(_SOLARIS) || defined(_HPUX)
    #include <unistd.h>
#endif
#if defined(WIN32)
void __declspec(dllexport) *bipGetMessageflowNodeFactory()
#else
CciFactory LilFactoryExportPrefix * LilFactoryExportSuffix bipGetMessageflowNodeFactory()
#endif
{
    static char *functionName="bipGetMessageflowNodeFactory()";
    void* factoryObject = NULL;
    CciChar *factoryName = C.TIMER_PLUGIN_FACTORY;
#if defined(POST_201_VERSION)
	CCI_EXCEPTION_ST exception_st={CCI_EXCEPTION_ST_DEFAULT};
#else
    CCI_EXCEPTION_ST exception_st;
#endif
    int rc;
#if defined(POST_201_VERSION)
	static CNI_VFT vftable={CNI_VFT_DEFAULT};
#else
    static CNI_VFT vftable;
#endif
	bool_t resourceCreated = TRUE;
	char *resourceName,
		 msgBuffer[200];

    TRACE_ENTRY("bipGetMessageflowNodeFactory", ("void"));
    currentPID = getpid();
    TRACE_MID(TRACE_INFO, ("<------------------------------------------------------>"));
    TRACE_MID(TRACE_INFO, ("<**** Start of timer deployment (pid=%d) ****>",
		                   currentPID));
    TRACE_MID(TRACE_INFO, ("<------------------------------------------------------>"));
	/* create the global syncronisation resources we need here */

	/* create required mutexes */
	if (resourceCreated)
	{
        TRACE_MID(TRACE_INFO, ("About to create timerChainMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &timerChainMutex) != THREAD_OK)
		{
			resourceName = "timerChainMutex";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create alarmThreadSyncMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &alarmThreadSyncMutex) != THREAD_OK)
		{
			resourceName = "alarmThreadSyncMutex";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create shmemInitTermMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &shmemInitTermMutex) != THREAD_OK)
		{
			resourceName = "shmemInitTermMutex";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create shmemThreadWorkMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &shmemThreadWorkMutex) != THREAD_OK)
		{
			resourceName = "shmemInitTermMutex";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create shmemThreadSyncMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &shmemThreadSyncMutex) != THREAD_OK)
		{
			resourceName = "shmemThreadSyncMutex";
			resourceCreated = FALSE;
		}
	}

	/* create required condition variables */
	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create alarmThreadWorkCondVar"));
		if (threadCreateCondVar(timerChainMutex,
								&alarmThreadWorkCondVar) != THREAD_OK)
		{
			resourceName = "alarmThreadWorkCondVar";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create alarmThreadSyncCondVar"));
		if (threadCreateCondVar(alarmThreadSyncMutex,
								&alarmThreadSyncCondVar) != THREAD_OK)
		{
			resourceName = "alarmThreadSyncCondVar";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create shmemThreadWorkCondVar"));
		if (threadCreateCondVar(shmemThreadWorkMutex,
								&shmemThreadWorkCondVar) != THREAD_OK)
		{
			resourceName = "shmemThreadWorkCondVar";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create shmemThreadSyncCondVar"));
		if (threadCreateCondVar(shmemThreadSyncMutex,
								&shmemThreadSyncCondVar) != THREAD_OK)
		{
			resourceName = "shmemThreadSyncCondVar";
			resourceCreated = FALSE;
		}
	}

	/* process any errors that have occurred */
	if (!resourceCreated)
	{
		sprintf(msgBuffer, "Timer Plugin Factory failed to create resource %s",
						   resourceName);
		/* generate an exception here... */
		cciThrowException(&rc, 
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_RESOURCE_CREATION_FAILED,  /* messageNumber */ 
			              msgBuffer, /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
			              "TimerPlugInFactory",
						  resourceName,
			              (char *)0);
	}


    /* Setup virtual function table with addresses of implementation functions */
#if !defined(POST_201_VERSION)
    memset(&vftable, 0, sizeof(vftable));
#endif
    vftable.iFpCreateNodeContext     = createNodeContext;
    vftable.iFpDeleteNodeContext     = deleteNodeContext;
    vftable.iFpGetAttributeName      = getAttributeName;
    vftable.iFpSetAttribute          = setAttribute;
    vftable.iFpGetAttribute          = getAttribute;
    vftable.iFpEvaluate              = evaluate;


    /* Create the Node Factory for this plug-in */
    factoryObject = cniCreateNodeFactory(&rc, factoryName);
    if (factoryObject == CCI_NULL_ADDR)
	{
        if (rc == CCI_EXCEPTION)
		{
            cciGetLastExceptionData(&rc, 
				                    &exception_st);
            /* Any local error handling may go here */
            cciRethrowLastException(&rc);
		}
        /* Any further local error handling can go here */
        return(factoryObject);
	}
    /* Define the TimerCreate and TimerCancel nodes supported by our factory */

	/*
	 * Define the TimerCreate node
	 */
    cniDefineNodeClass(&rc, 
			           factoryObject,
					   C.TIMER_CREATE_NODE,
					   &vftable); 
    if (rc != CCI_SUCCESS)
	{
        TRACE_MID(TRACE_ERROR, ("Failed to create TimerCreate node "
			                    "(cniDefineNodeClass rc=%d)", rc));
		/* generate an exception here... */
		cciThrowException(&rc,
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_NODE_CLASS_CREATION_FAILED,  /* messageNumber */ 
			              "Timer Plugin Factory failed to create TimerCreate node", /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
			              "TimerPlugInFactory",
			              TIMER_CREATE_NODE_MB,
			              (char *)0);
	}

	/*
	 * Define the TimerCancel node
	 */
    cniDefineNodeClass(&rc, 
			           factoryObject, 
					   C.TIMER_CANCEL_NODE, 
					   &vftable); 
    if (rc != CCI_SUCCESS)
	{
        TRACE_MID(TRACE_ERROR, ("Failed to create TimerCancel node "
			                    "(cniDefineNodeClass rc=%d)", rc));
		/* generate an exception here... */
		cciThrowException(&rc,
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_NODE_CLASS_CREATION_FAILED,  /* messageNumber */ 
			              "Timer Plugin Factory failed to create TimerCancel node", /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
			              "TimerPlugInFactory",
			              TIMER_CANCEL_NODE_MB,
			              (char *)0);
	}

    /* Log a success message */
    TRACE_MID(TRACE_INFO, ("About to issue cciLog success message"));
    cciLog(&rc,                       /* returnCode */
	       CCI_LOG_INFORMATION,       /* type */
		   __FILE__,                  /* file */
		   __LINE__,                  /* line */
		   functionName,              /* function */
		   C.PLUGIN_MESSAGE_CATALOGUE,  /* messageSource */
#if defined(WIN32)
           MJB_NODE_FACTORY_CREATED,  /* messageNumber */ 
#else
           MJB_NODE_FACTORY_CREATED_UNIX,  /* messageNumber */ 
#endif
		   "Timer Plugin Factory loading", /* traceText */
		   "TimerPlugInFactory",
		   (char *)0);

    /* Return address of this factory object to the broker */
    TRACE_RETURN_TYPE(RSN_021, void *, factoryObject);
}


/* local functions */


/******************************************************************************/
/* attachSharedMemoryIfRequired                                               */
/******************************************************************************/
int attachSharedMemoryIfRequired(IN context_t *contextP,
								 IN scope_t scope)
{
	int shmSize,
		retCode;
	globalCB_t dummyCB = {GLOBAL_CB_DEFAULT};
	executionGroupEntry_t *currentEGP;
    static char *functionName="attachSharedMemoryIfRequired";
	char ownerKey[MAX_STRING_SIZE + 1],
	     segNameKey[MAX_STRING_SIZE + 1];
	bool_t scopeIsBroker;
#if !defined(WIN32)
	unsigned long *numericKeyP;
	int	i;
#endif

    TRACE_ENTRY("attachSharedMemoryIfRequired", ("contextP=%.8x, scope=%s", 
		                                          contextP,
												  scope==SCOPE_BROKER ? "Broker" :
	                                              "Platform"));

    scopeIsBroker = (scope == SCOPE_BROKER);

	TRACE_MID(TRACE_INFO, ("About to lock shared memory mutex"));
	threadMutexLock(shmemInitTermMutex);
	if (!sharedMemoryAttached_Broker && scope==SCOPE_BROKER ||
	    !sharedMemoryAttached_Platform && scope==SCOPE_PLATFORM)
	{
		shmSize = getSharedSegmentSize(sharedSegmentEnvironmentVariableName,
							           SHARED_SEGMENT_DEFAULT_KSIZE * 1024);
#if defined(WIN32)
		strcpy(segNameKey, SHARED_MEMORY_SEGMENT_NAME_PREFIX);
		if (scope == SCOPE_BROKER)
		{
			strcat(segNameKey, brokerName);
		}
		else
		{
			strcat(segNameKey, "Platform");
		}
		TRACE_MID(TRACE_INFO, ("Segment name key is: %s", segNameKey));
#else
		strcpy(segNameKey, SHARED_SEGMENT_KEY);
		numericKeyP = (unsigned long *)segNameKey;
		TRACE_MID(TRACE_INFO, ("*numericKeyP value is %.8x", *numericKeyP));
		if (scope == SCOPE_BROKER)
		{
			// hash the broker name and incorporate into the segment key
			// need to find a good hashing algorithm to replace this temporary one
			for (i=0; i<(int)strlen(brokerName); i++)
			{
				*numericKeyP ^= (int)brokerName[i] << 8;
				*numericKeyP += (int)brokerName[i];
				*numericKeyP <<= 1;
				TRACE_MID(TRACE_INFO, ("*numericKeyP value is %.8x", *numericKeyP));
			}
		}
#endif

		/* build owner key for broker scope */
		if (scope == SCOPE_BROKER)
		{
			strcpy(ownerKey, SHARED_SEGMENT_BROKER_SCOPE_KEY_PREFIX);
			strcat(ownerKey, brokerName);
			TRACE_MID(TRACE_INFO, ("Owner key is: %s", ownerKey));
		}
		TRACE_MID(TRACE_INFO, ("About to attempt to request/attach shared "
			                   "memory segment '%s' of %d bytes",
							    segNameKey,
								shmSize));

		retCode = shmAttach(segNameKey,
							shmSize,
							(void *)&dummyCB,
							sizeof(dummyCB),
							currentPID,
							scope==SCOPE_BROKER ? brokerName : SHARED_SEGMENT_PLATFORM_SCOPE_KEY,
							scope==SCOPE_BROKER);

		if (retCode != TT_SUCCESS)
		{
			/* deal with error, including unlocking mutex */
			TRACE_MID(TRACE_ERROR, ("Failed to attach shared memory"));

			TRACE_MID(TRACE_INFO, ("About to unlock shared memory mutex"));
			threadMutexUnlock(shmemInitTermMutex);
			TRACE_RETURN(RSN_001, TT_FAILED_TO_ATTACH_GLOBAL_STORAGE);
		}

		if (scope == SCOPE_BROKER)
		{
			sharedMemoryCBP_Broker = (globalCB_t *)shmGetCBAddress(TRUE);
			sharedMemoryAttached_Broker = TRUE;
		}
		else
		{
			sharedMemoryCBP_Platform = (globalCB_t *)shmGetCBAddress(FALSE);
			sharedMemoryAttached_Platform = TRUE;
		}
	}

	/* 
	 * Tidy up shared memory if any previous instance of this execution group failed without
	 * properly removing its resources.
	 */

	/* lock shared memory */
	shmLock(scopeIsBroker);

	// test locking...sleep for a while and wait for other process to try lock
    // TRACE_MID(TRACE_INFO, ("Sleeping to test global lock"));
    // Sleep(20000);  // sleep for 20 seconds

	currentEGP = findExecutionGroupEntry(scopeIsBroker);

    if (currentEGP)
	{
		removeEGTimersFromGlobalStorage(currentEGP,
			                            scopeIsBroker);

		/* remove stale pid from pid list */
		retCode = shmRemovePidFromList(currentEGP->pid,
					                   TRUE,  // we've got memory locked
									   scopeIsBroker,  // Broker scope
				   	                   NULL);
	}

	/* 
	 * If an execution group entry already exists, then store our pid in it, otherwise create 
	 * a new entry.
	 */
    if (!currentEGP)
	{
		currentEGP = createNewExecutionGroupEntry(scopeIsBroker);
	}

	/* store our pid in the entry */
	currentEGP->pid = currentPID;

	/* unlock shared memory */
	shmUnlock(scopeIsBroker);

	/* 
	 * We also need to start a background thread to do garbage collection for the areas
	 * we allocate in the shared memory space.
	 */
	if (!shmemThreadCreated)
	{
		retCode = createSharedMemoryGarbageCollectionThread(contextP);

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_INFO, ("About to unlock shared memory mutex"));
			threadMutexUnlock(shmemInitTermMutex);
			TRACE_RETURN(RSN_005, TT_FAILED_TO_CREATE_GLOBAL_STORAGE_GARBAGE_COLLECTOR);
		}
	}

	TRACE_MID(TRACE_INFO, ("About to unlock shared memory mutex"));
	threadMutexUnlock(shmemInitTermMutex);

	TRACE_RETURN(RSN_001, TT_SUCCESS);
}


/******************************************************************************/
/* findExecutionGroupEntry                                                    */
/******************************************************************************/
executionGroupEntry_t *findExecutionGroupEntry(IN bool_t scopeIsBroker)
{
    executionGroupEntry_t *currentEGP = NULL;
	char fullyQualifiedName[MAX_STRING_SIZE * 2 + 1 + 1];
    TRACE_ENTRY("findExecutionGroupEntry", ("scopeIsBroker=%s", 
		                                     scopeIsBroker ? "TRUE" : "FALSE"));

	/* on entry, assume caller has shared memory locked */

	/* see if this execution group already exists */
	currentEGP = SHM_HANDLE2ADDRESS((scopeIsBroker ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->executionGroupAnchor.nextEntry, 
		                            executionGroupEntry_t *, 
									scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

	/* build the fully qualified EG name by prepending with broker name and underscore */
	strcpy(fullyQualifiedName, brokerName);
	strcat(fullyQualifiedName, "_");
	strcat(fullyQualifiedName, executionGroupName);

	TRACE_MID(TRACE_INFO, ("Looking for execution group entry with fully qualified name '%s'",
			                fullyQualifiedName));
	while(currentEGP)
	{
		TRACE_MID(TRACE_INFO, ("Processing entry at %.8x (handle=%.8x)",
			                   currentEGP, SHM_ADDRESS2HANDLE(currentEGP,
							   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
		if (!strcmp(fullyQualifiedName,
			        SHM_HANDLE2ADDRESS(currentEGP->name, char *,
					                   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)))
		{
			TRACE_MID(TRACE_INFO, ("Found a match containing pid %d", currentEGP->pid));
			break;
		}

		/* address the next in the list */
		currentEGP = SHM_HANDLE2ADDRESS(currentEGP->nextEntry, executionGroupEntry_t *,
			                            scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	}

	TRACE_RETURN_TYPE(RSN_001, executionGroupEntry_t *, currentEGP);
}


/******************************************************************************/
/* removeEGTimersFromGlobalStorage                               */
/******************************************************************************/
void removeEGTimersFromGlobalStorage(IN executionGroupEntry_t *currentEGP,
									 IN bool_t scopeIsBroker)
{
	globalTimerInstance_t *currentGlobalInstanceP,
			 			  *nextGlobalInstanceP;
    TRACE_ENTRY("removeEGTimersFromGlobalStorage", ("currentEGP=%.8x, scope=%s",
                                                     currentEGP,
													 scopeIsBroker ? "TRUE" : "FALSE"));

	/* run down global chains and remove any timers owned by dead pid */
	currentGlobalInstanceP = 
		SHM_HANDLE2ADDRESS((scopeIsBroker ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalMsgIdP,
			               globalTimerInstance_t *,
						   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

	TRACE_MID(TRACE_INFO, ("Checking global MsgId chain"));
	while(currentGlobalInstanceP)
	{
		TRACE_MID(TRACE_INFO, ("Checking for dead global entry at %.8x (handle=%.8x",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
		nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalMsgIdP, 
				                                 globalTimerInstance_t *,
												 scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		if (currentGlobalInstanceP->executionGroupPID = currentEGP->pid)
		{
			TRACE_MID(TRACE_INFO, ("Freeing dead timer at %.8x (handle=%.8x)",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
			/* remove from global chains and free global storage */
			unchainTimer((timerInstance_t *)currentGlobalInstanceP); 
			shmFree(currentGlobalInstanceP, TRUE, currentGlobalInstanceP->scope==SCOPE_BROKER);
		}
		currentGlobalInstanceP = nextGlobalInstanceP;
	}

	currentGlobalInstanceP = 
		SHM_HANDLE2ADDRESS((scopeIsBroker ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalCorrelIdP,
			               globalTimerInstance_t *, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Checking global CorrelId chain"));
	while(currentGlobalInstanceP)
	{
		TRACE_MID(TRACE_INFO, ("Checking for dead global entry at %.8x (handle=%.8x",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
		nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalCorrelIdP, 
				                                 globalTimerInstance_t *,
												 scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		if (currentGlobalInstanceP->executionGroupPID = currentEGP->pid)
		{
			TRACE_MID(TRACE_INFO, ("Freeing dead timer at %.8x (handle=%.8x)",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
			/* remove from global chains and free global storage */
			unchainTimer((timerInstance_t *)currentGlobalInstanceP);
			shmFree(currentGlobalInstanceP, TRUE, currentGlobalInstanceP->scope==SCOPE_BROKER);
		}
		currentGlobalInstanceP = nextGlobalInstanceP;
	}

	currentGlobalInstanceP = 
		SHM_HANDLE2ADDRESS((scopeIsBroker ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalBothP,
			               globalTimerInstance_t *, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Checking global Both chain"));
	while(currentGlobalInstanceP)
	{
		TRACE_MID(TRACE_INFO, ("Checking for dead global entry at %.8x (handle=%.8x",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
		nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalBothP, 
				                                 globalTimerInstance_t *, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		if (currentGlobalInstanceP->executionGroupPID = currentEGP->pid)
		{
			TRACE_MID(TRACE_INFO, ("Freeing dead timer at %.8x (handle=%.8x)",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
			/* remove from global chains and free global storage */
			unchainTimer((timerInstance_t *)currentGlobalInstanceP);
			shmFree(currentGlobalInstanceP, TRUE, currentGlobalInstanceP->scope==SCOPE_BROKER);
		}
		currentGlobalInstanceP = nextGlobalInstanceP;
	}

	/* check the dead timer chain also */
	currentGlobalInstanceP = 
		SHM_HANDLE2ADDRESS((scopeIsBroker ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalDeadInstanceAnchor.nextGlobalP,
			               globalTimerInstance_t *, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Checking global dead timer chain"));
	while(currentGlobalInstanceP)
	{
		TRACE_MID(TRACE_INFO, ("Checking for dead global entry at %.8x (handle=%.8x",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
		nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalP, 
				                                 globalTimerInstance_t *,
												 scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		if (currentGlobalInstanceP->executionGroupPID = currentEGP->pid)
		{
			TRACE_MID(TRACE_INFO, ("Freeing dead timer at %.8x (handle=%.8x)",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
			/* remove from dead chain and free global storage */
			REMOVE_FROM_GLOBAL_TIMER_CHAIN(currentGlobalInstanceP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			shmFree(currentGlobalInstanceP, TRUE, currentGlobalInstanceP->scope==SCOPE_BROKER);
		}
		currentGlobalInstanceP = nextGlobalInstanceP;
	}

	TRACE_RETURN_VOID(RSN_001);
}


/******************************************************************************/
/* createNewExecutionGroupEntry                                               */
/******************************************************************************/
executionGroupEntry_t *createNewExecutionGroupEntry(IN bool_t scopeIsBroker)
{
	int retCode;
	char *egName;
	char fullyQualifiedName[MAX_STRING_SIZE * 2 + 1 + 1];
    static char *functionName="createNewExecutionGroupEntry";
	executionGroupEntry_t defaultEGEntry = {EXECUTION_GROUP_ENTRY_DEFAULT},
		                  *anchorEGP,
		                  *currentEGP;
    TRACE_ENTRY("createNewExecutionGroupEntry", ("scopeIsBroker=%s",
												  scopeIsBroker ? "TRUE" : "FALSE"));

	/* build the fully qualified EG name by prepending with broker name and underscore */
	strcpy(fullyQualifiedName, brokerName);
	strcat(fullyQualifiedName, "_");
	strcat(fullyQualifiedName, executionGroupName);

	TRACE_MID(TRACE_INFO, ("Fully qualified name is '%s'",
			                fullyQualifiedName));

	currentEGP = (executionGroupEntry_t *)shmMalloc(sizeof(executionGroupEntry_t),
			                                        TRUE,
						                            scopeIsBroker,
						                            &retCode);
	CHECK_SHMEM_ERROR(retCode, 
		              functionName, 
					  scopeIsBroker, 
			          currentEGP, 
					  executionGroupEntry_t *, 
					  sizeof(executionGroupEntry_t),
				      TRUE,   /* lock is already held */
				      platformIsUnix,
					  FALSE); // gc thread doesn't exist yet, so retry=FALSE  

	TRACE_MID(TRACE_INFO, ("New execution group entry block allocated at %.8x (handle=%.8x)",
			               currentEGP, SHM_ADDRESS2HANDLE(currentEGP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));

	/* initialise control block to default values */
    memcpy(currentEGP, &defaultEGEntry, sizeof(executionGroupEntry_t));

	egName = (char *)shmMalloc(strlen(fullyQualifiedName) + 1,  // allow for null
			                   TRUE,
					           scopeIsBroker,
					           &retCode);

	CHECK_SHMEM_ERROR(retCode, 
		              functionName, 
					  scopeIsBroker, 
			          egName, 
					  char *, 
					  strlen(fullyQualifiedName) + 1,
				      TRUE,   /* lock is already held */
				      platformIsUnix,  
					  FALSE); // gc thread doesn't exist yet, so retry=FALSE  

	TRACE_MID(TRACE_INFO, ("New execution group name buffer allocated at %.8x (handle=%.8x)",
			               egName, SHM_ADDRESS2HANDLE(egName, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));

	/* copy the EG name into the buffer */
	strcpy(egName, fullyQualifiedName);

	/* store the name in the control block as a handle value */
	currentEGP->name = SHM_ADDRESS2HANDLE(egName, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Name handle %.8x stored in execution group entry", 
		                    currentEGP->name));

	/* add the entry to the global chain */
	anchorEGP = &(scopeIsBroker ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->executionGroupAnchor;
	TRACE_MID(TRACE_INFO, ("About to chain entry to anchor at %.8x", anchorEGP));
	INSERT_AFTER_IN_GLOBAL_EG_NAME_CHAIN(anchorEGP, currentEGP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("New execution group entry added to global list"));

	TRACE_RETURN_TYPE(RSN_001, executionGroupEntry_t *, currentEGP);
}

/******************************************************************************/
/* removeExecutionGroupFromGlobalStorage                                      */
/******************************************************************************/
void removeExecutionGroupFromGlobalStorage(IN bool_t scopeIsBroker)
{
	executionGroupEntry_t *currentEGP = NULL;
    TRACE_ENTRY("removeExecutionGroupFromGlobalStorage", ("scopeIsBroker=%s",
												           scopeIsBroker ? "TRUE" : "FALSE"));

	/* on entry, assume caller has shared memory locked */
	currentEGP = findExecutionGroupEntry(scopeIsBroker);
	if (!currentEGP)
	{
		TRACE_MID(TRACE_INFO, ("No execution group found"));
		TRACE_RETURN_VOID(RSN_001);
	}
	TRACE_MID(TRACE_INFO, ("Execution group entry located at %.8x", currentEGP));
    
	/* unchain execution group entry */
	REMOVE_FROM_GLOBAL_EG_NAME_CHAIN(currentEGP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Execution group entry removed from global chains"));

	/* free the global storage */
	if (currentEGP->name != SHM_NULL_HANDLE)
	{
		shmFree(SHM_HANDLE2ADDRESS(currentEGP->name, void *, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform),
 			    TRUE,  // global lock already taken
				scopeIsBroker);
		TRACE_MID(TRACE_INFO, ("Execution group name buffer freed"));
	}
	TRACE_MID(TRACE_INFO, ("Execution group entry control block freed"));

	TRACE_RETURN_VOID(RSN_001);
}


